# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A few helpers for luigi-tools test suite."""
from pathlib import Path


def check_existing_file(filename):
    """Check if a file exists."""
    return Path(filename).exists()


def create_empty_file(filename):
    """Create an empty file."""
    with open(filename, "w", encoding="utf-8"):
        pass


def check_empty_file(filename):
    """Check that a file is empty."""
    with open(filename, encoding="utf-8") as f:
        return f.read() == ""


def create_not_empty_file(filename):
    """Create a not empty file."""
    with open(filename, "w", encoding="utf-8") as f:
        f.write("NOT EMPTY")


def check_not_empty_file(filename):
    """Check that a file is not empty."""
    with open(filename, encoding="utf-8") as f:
        return f.read() == "NOT EMPTY"



"""Extension to define a new Sphinx directive for objects moved to the `luigi` package."""
from docutils import nodes
from docutils.parsers.rst import directives
from docutils.statemachine import StringList
from sphinx.locale import _
from sphinx.util.docutils import SphinxDirective


class luigi_moved(nodes.Admonition, nodes.Element):
    """Base node for objects moved to the `luigi` package."""


def visit_moved_node(self, node):
    """Visit node."""
    self.visit_admonition(node)


def depart_moved_node(self, node):
    """Depart node."""
    self.depart_admonition(node)


class MovedDirective(SphinxDirective):
    """New Sphinx directive for objects moved to the `luigi` package."""

    has_content = True
    option_spec = {
        "luigi_version": directives.unchanged,
        "previous_luigi_version": directives.unchanged,
        "deprecation_version": directives.unchanged,
    }

    def run(self):
        """Actual run method."""
        targetid = "luigi_moved-%d" % self.env.new_serialno("luigi_moved")
        targetnode = nodes.target("", "", ids=[targetid])

        msg = "This feature was moved to the :mod:`luigi` package"
        if ("luigi_version" in self.options) == ("previous_luigi_version" in self.options):
            raise ValueError(
                "Either the 'luigi_version' or the 'previous_luigi_version' argument of the "
                f"'{self.name}' directive must be not None but not both of them"
            )
        if "luigi_version" in self.options:
            msg += f" and is available since version {self.options['luigi_version']}."
        else:
            msg += (
                " and will be available after the version "
                f"{self.options['previous_luigi_version']}."
            )
        body = [msg]

        if "deprecation_version" in self.options:
            body.append(f"It will be deprecated in version {self.options['deprecation_version']}.")
        else:
            body.append("It will be deprecated in a future version.")

        if self.content:
            body.append("")
            body.extend(self.content.data)
        todo_node = luigi_moved(body)
        todo_node += nodes.title(_("Moved to luigi"), _("Moved to luigi"))
        self.state.nested_parse(StringList(body), self.content_offset, todo_node)

        return [targetnode, todo_node]


def setup(app):
    """Connect to Sphinx setup."""
    app.add_node(
        luigi_moved,
        html=(visit_moved_node, depart_moved_node),
        latex=(visit_moved_node, depart_moved_node),
        text=(visit_moved_node, depart_moved_node),
    )
    app.add_directive("moved_to_luigi", MovedDirective)
    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration for the pytest test suite."""

# pylint: disable=empty-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
import os

import luigi
import pytest

import luigi_tools.task
from luigi_tools.util import set_luigi_config

from .tools import create_not_empty_file


@pytest.fixture(scope="function")
def tmp_working_dir(tmp_path):
    """Change working directory before a test and change it back when the test is finished."""
    cwd = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(cwd)


@pytest.fixture
def luigi_tools_params():
    """Simple luigi parameters for the TaskA task."""
    return {"TaskA": {"a_cfg": "default_value_in_cfg"}}


@pytest.fixture
def luigi_tools_working_directory(tmp_working_dir, luigi_tools_params):
    """Set luigi config."""
    with set_luigi_config(luigi_tools_params):
        yield tmp_working_dir


@pytest.fixture
def task_collection(tmpdir):
    """A collection of test classes."""

    class TaskClasses:
        """Class with some luigi tasks to test."""

        def __init__(self):
            self.tmpdir = tmpdir
            self.reset_classes()
            self.classes = self._classes()
            self.targets = self._targets()
            self.reset_classes()  # Reset again to return classes that are not registered by luigi

        def reset_classes(self):
            """Reset all test classes."""

            class TaskA(luigi_tools.task.WorkflowTask):
                """A simple test task."""

                counter = luigi.IntParameter(default=0)

                def run(self):
                    for i in luigi.task.flatten(self.output()):
                        create_not_empty_file(i.path)

                def output(self):
                    return luigi.LocalTarget(tmpdir / "TaskA.target")

            class TaskB(luigi_tools.task.WorkflowTask):
                """A simple test task."""

                def requires(self):
                    return TaskA()

                def run(self):
                    for i in luigi.task.flatten(self.output()):
                        create_not_empty_file(i.path)

                def output(self):
                    return [
                        luigi.LocalTarget(tmpdir / "TaskB.target"),
                        [
                            luigi.LocalTarget(tmpdir / "TaskB2.target"),
                            luigi.LocalTarget(tmpdir / "TaskB3.target"),
                        ],
                    ]

            class TaskC(luigi_tools.task.WorkflowTask):
                """A simple test task."""

                def requires(self):
                    return TaskA()

                def run(self):
                    for i in luigi.task.flatten(self.output()):
                        create_not_empty_file(i.path)

                def output(self):
                    return {
                        "first_target": luigi.LocalTarget(tmpdir / "TaskC.target"),
                        "second_target": luigi.LocalTarget(tmpdir / "TaskC2.target"),
                    }

            class TaskD(luigi_tools.task.WorkflowTask):
                """A simple test task."""

                def requires(self):
                    return [TaskB(), TaskC()]

                def run(self):
                    for i in luigi.task.flatten(self.output()):
                        create_not_empty_file(i.path)

                def output(self):
                    return [
                        luigi.LocalTarget(tmpdir / "TaskD.target"),
                        luigi.LocalTarget(tmpdir / "TaskD2.target"),
                    ]

            class TaskE(luigi_tools.task.WorkflowTask):
                """A simple test task."""

                def requires(self):
                    return TaskD()

                def run(self):
                    for i in luigi.task.flatten(self.output()):
                        create_not_empty_file(i.path)

                def output(self):
                    return {
                        "first_target": luigi.LocalTarget(tmpdir / "TaskE.target"),
                        "other_targets": {
                            "second_target": luigi.LocalTarget(tmpdir / "TaskE2.target"),
                            "third_target": luigi.LocalTarget(tmpdir / "TaskE3.target"),
                        },
                    }

            self.TaskA = TaskA
            self.TaskB = TaskB
            self.TaskC = TaskC
            self.TaskD = TaskD
            self.TaskE = TaskE

        def _classes(self):
            return [
                self.TaskA,
                self.TaskB,
                self.TaskC,
                self.TaskD,
                self.TaskE,
            ]

        def _targets(self):
            targets = {}
            for task in self.classes:
                targets[task.__name__] = task().output()
            return targets

    return TaskClasses()



# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for luigi-tools parameters."""

import collections
import dataclasses
import json
import typing

# pylint: disable=empty-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument
import warnings

import luigi
import mock
import pytest
from luigi.freezing import FrozenOrderedDict

import luigi_tools.parameter
import luigi_tools.target
import luigi_tools.task
import luigi_tools.util
from luigi_tools.util import set_luigi_config

from .tools import create_empty_file


def test_ext_parameter(luigi_tools_working_directory):
    """Test the `luigi_tools.parameter.ExtParameter` class."""

    class TaskExtParameter(luigi.Task):
        """A simple test task."""

        a = luigi_tools.parameter.ExtParameter(default=".ext_from_default")
        test_type = luigi.Parameter(default="default")

        def run(self):
            if self.test_type == "default":
                assert self.a == "ext_from_default"
            else:
                assert self.a == "ext_from_cfg"
            create_empty_file(self.output().path)

        def output(self):
            return luigi.LocalTarget(
                luigi_tools_working_directory / f"test_ext_parameter_{self.a}_{self.test_type}.test"
            )

    assert luigi.build([TaskExtParameter(test_type="default")], local_scheduler=True)

    with set_luigi_config({"TaskExtParameter": {"a": "ext_from_cfg"}}):
        assert luigi.build([TaskExtParameter(test_type="cfg")], local_scheduler=True)

    with set_luigi_config({"TaskExtParameter": {"a": ".ext_from_cfg"}}):
        assert luigi.build([TaskExtParameter(test_type="cfg_with_dot")], local_scheduler=True)


def test_ratio_parameter(tmpdir):
    """Test the `luigi_tools.parameter.RatioParameter` class."""

    class TaskRatioParameter(luigi.Task):
        """A simple test task."""

        a = luigi_tools.parameter.RatioParameter(default=0.5)
        b = luigi.FloatParameter(default=0.5)

        def run(self):
            assert self.a == self.b
            create_empty_file(self.output().path)

        def output(self):
            return luigi.LocalTarget(tmpdir / f"test_ratio_parameter_{self.a}_{self.b}.test")

    assert luigi.build([TaskRatioParameter()], local_scheduler=True)
    assert luigi.build([TaskRatioParameter(a=0.25, b=0.25)], local_scheduler=True)
    assert luigi.build([TaskRatioParameter(a=0, b=0)], local_scheduler=True)
    assert luigi.build([TaskRatioParameter(a=1, b=1)], local_scheduler=True)
    with pytest.raises(ValueError):
        assert luigi.build([TaskRatioParameter(a=-1, b=-1)], local_scheduler=True)
    with pytest.raises(ValueError):
        assert luigi.build([TaskRatioParameter(a=2, b=2)], local_scheduler=True)
    with pytest.raises(ValueError):
        assert luigi.build([TaskRatioParameter(a="a", b="a")], local_scheduler=True)


class TestOptionalParameter:
    """Test the `luigi_tools.parameter.Optional*` classes.

    .. note::

        These classes were ported to the luigi package but the tests were kept here for example
        purpose.
    """

    def test_optional_parameter(self, luigi_tools_working_directory):
        """Simple test for optional parameters."""

        # pylint: disable=too-many-statements
        class TaskFactory:
            """A class factory."""

            def __call__(self):
                class TaskOptionalParameter(luigi.Task):
                    """A simple test task."""

                    a = luigi.parameter.OptionalIntParameter(default=1)
                    b = luigi.parameter.OptionalFloatParameter(default=1.5)
                    c = luigi.parameter.OptionalNumericalParameter(
                        default=0.75, min_value=0, max_value=1, var_type=float
                    )
                    d = luigi_tools.parameter.OptionalRatioParameter(default=0.5)
                    e = luigi.parameter.OptionalIntParameter(default=None)
                    f = luigi.parameter.OptionalListParameter(default=None)
                    g = luigi.parameter.OptionalChoiceParameter(default=None, choices=["a", "b"])
                    h = luigi.parameter.OptionalStrParameter(default="h")
                    i = luigi_tools.parameter.OptionalBoolParameter(default=None)
                    expected_a = luigi.IntParameter(default=1)
                    expected_b = luigi.FloatParameter(default=1.5)
                    expected_c = luigi.NumericalParameter(
                        default=0.75, min_value=0, max_value=1, var_type=float
                    )
                    expected_d = luigi_tools.parameter.RatioParameter(default=0.5)
                    expected_e = luigi.IntParameter(default=None)
                    expected_f = luigi.ListParameter(default=None)
                    expected_g = luigi.ChoiceParameter(default="null", choices=["a", "b", "null"])
                    expected_h = luigi.Parameter(default="h")
                    expected_i = luigi.BoolParameter(default=None)

                    def run(self):
                        if self.expected_g == "null":
                            self.expected_g = None
                        print(
                            "self.a =",
                            self.a,
                            "self.b =",
                            self.b,
                            "self.c =",
                            self.c,
                            "self.d =",
                            self.d,
                            "self.e =",
                            self.e,
                            "self.f =",
                            self.f,
                            "self.g =",
                            self.g,
                            "self.h =",
                            self.h,
                            "self.i =",
                            self.i,
                        )
                        assert self.a == self.expected_a
                        assert self.b == self.expected_b
                        assert self.c == self.expected_c
                        assert self.d == self.expected_d
                        assert self.e == self.expected_e
                        assert self.f == self.expected_f
                        assert self.g == self.expected_g
                        assert self.h == self.expected_h
                        assert self.i == self.expected_i
                        create_empty_file(self.output().path)

                    def output(self):
                        # pylint: disable=consider-using-f-string
                        return luigi.LocalTarget(
                            luigi_tools_working_directory
                            / (
                                "test_optional_parameter_"
                                "{}_{}_{}_{}_{}_{}_{}_{}_{}.test".format(
                                    self.a,
                                    self.b,
                                    self.c,
                                    self.d,
                                    self.e,
                                    self.f,
                                    self.g,
                                    self.h,
                                    self.i,
                                )
                            )
                        )

                return TaskOptionalParameter

        factory = TaskFactory()

        task = factory()
        assert luigi.build([task()], local_scheduler=True)

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "a": "",
                    "b": "",
                    "c": "",
                    "d": "",
                    "e": "",
                    "f": "",
                    "g": "",
                    "h": "",
                    "i": "",
                }
            }
        ):
            task = factory()
            assert luigi.build(
                [
                    task(
                        expected_a=None,
                        expected_b=None,
                        expected_c=None,
                        expected_d=None,
                        expected_e=None,
                        expected_f=None,
                        expected_g="null",
                        expected_h=None,
                        expected_i=None,
                    )
                ],
                local_scheduler=True,
            )

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "a": "0",
                    "b": "0.7",
                    "c": "0.8",
                    "d": "0.9",
                    "e": "1",
                    "f": "[1, 2]",
                    "g": "b",
                    "h": "b",
                    "i": "true",
                }
            }
        ):
            task = factory()
            assert luigi.build(
                [
                    task(
                        expected_a=0,
                        expected_b=0.7,
                        expected_c=0.8,
                        expected_d=0.9,
                        expected_e=1,
                        expected_f=[1, 2],
                        expected_g="b",
                        expected_h="b",
                        expected_i=True,
                    )
                ],
                local_scheduler=True,
            )

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "a": "not numerical",
                }
            }
        ):
            task = factory()
            with pytest.raises(ValueError):
                assert luigi.build([task()], local_scheduler=True)

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "b": "not numerical",
                }
            }
        ):
            task = factory()
            with pytest.raises(ValueError):
                assert luigi.build([task()], local_scheduler=True)

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "c": "not numerical",
                }
            }
        ):
            task = factory()
            with pytest.raises(ValueError):
                assert luigi.build([task()], local_scheduler=True)

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "d": "not numerical",
                }
            }
        ):
            task = factory()
            with pytest.raises(ValueError):
                assert luigi.build([task()], local_scheduler=True)

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "f": "not list",
                }
            }
        ):
            task = factory()
            with pytest.raises(ValueError):
                assert luigi.build([task()], local_scheduler=True)

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "g": "not dict",
                }
            }
        ):
            task = factory()
            with pytest.raises(ValueError):
                assert luigi.build([task()], local_scheduler=True)

        with set_luigi_config(
            {
                "TaskOptionalParameter": {
                    "i": "not bool",
                }
            }
        ):
            task = factory()
            with pytest.raises(ValueError):
                assert luigi.build([task()], local_scheduler=True)

        class TaskOptionalParameterWarning(luigi.Task):
            """A simple test task."""

            a = luigi.parameter.OptionalIntParameter(default=1)
            expected_a = luigi.IntParameter(default=1)

            def run(self):
                print("self.a =", self.a)
                assert self.a == self.expected_a
                create_empty_file(self.output().path)

            def output(self):
                return luigi.LocalTarget(
                    luigi_tools_working_directory / f"test_optional_parameter_warning_{self.a}.test"
                )

        with warnings.catch_warnings(record=True) as w:
            warnings.filterwarnings(
                action="ignore",
                category=Warning,
            )
            warnings.simplefilter(
                action="always",
                category=luigi.parameter.OptionalParameterTypeWarning,
            )
            assert luigi.build(
                [TaskOptionalParameterWarning(a="zz", expected_a="zz")],
                local_scheduler=True,
            )

            assert len(w) == 1
            assert issubclass(w[0].category, luigi.parameter.OptionalParameterTypeWarning)
            assert str(w[0].message) == (
                """OptionalIntParameter "a" with value "zz" is not of type "int" or None."""
            )

    def test_warning(self):
        """Test that warnings are properly emitted."""

        class TestOptionalFloatParameterSingleType(
            luigi.parameter.OptionalParameterMixin, luigi.FloatParameter
        ):
            """Class to parse optional float parameters."""

            expected_type = float

        class TestOptionalFloatParameterMultiTypes(
            luigi.parameter.OptionalParameterMixin, luigi.FloatParameter
        ):
            """Class to parse optional float parameters."""

            expected_type = (int, float)

        class TestConfig(luigi.Config):
            """A simple config task."""

            param_single = TestOptionalFloatParameterSingleType()
            param_multi = TestOptionalFloatParameterMultiTypes()

        with warnings.catch_warnings(record=True) as record:
            TestConfig(param_single=0.0, param_multi=1.0)

        assert len(record) == 0

        with warnings.catch_warnings(record=True) as record:
            warnings.filterwarnings(
                action="ignore",
                category=Warning,
            )
            warnings.simplefilter(
                action="always",
                category=luigi.parameter.OptionalParameterTypeWarning,
            )
            assert luigi.build(
                [TestConfig(param_single="0", param_multi="1")], local_scheduler=True
            )

        assert len(record) == 2
        assert issubclass(record[0].category, luigi.parameter.OptionalParameterTypeWarning)
        assert issubclass(record[1].category, luigi.parameter.OptionalParameterTypeWarning)
        assert str(record[0].message) == (
            """TestOptionalFloatParameterSingleType "param_single" with value "0" is not of type """
            """"float" or None."""
        )
        assert str(record[1].message) == (
            """TestOptionalFloatParameterMultiTypes "param_multi" with value "1" is not of any """
            """type in ["int", "float"] or None."""
        )

    def actual_test(self, cls, default, expected_value, expected_type, bad_data, **kwargs):
        """A test function used for several parameter types."""

        class TestConfig(luigi.Config):
            """A simple config task."""

            param = cls(default=default, **kwargs)
            empty_param = cls(default=default, **kwargs)

            def run(self):
                assert self.param == expected_value
                assert self.empty_param is None

        # Test parsing empty string (should be None)
        assert cls(**kwargs).parse("") is None

        # Test that warning is raised only with bad type
        with mock.patch("luigi.parameter.warnings") as warnings:
            TestConfig()
            warnings.warn.assert_not_called()

        if cls != luigi.parameter.OptionalChoiceParameter:
            with mock.patch("luigi.parameter.warnings") as warnings:
                TestConfig(param=None)
                warnings.warn.assert_not_called()

            with mock.patch("luigi.parameter.warnings") as warnings:
                TestConfig(param=bad_data)
                if cls == luigi_tools.parameter.OptionalBoolParameter:
                    warnings.warn.assert_not_called()
                else:
                    assert warnings.warn.call_count == 1
                    # pylint: disable=consider-using-f-string
                    warnings.warn.assert_called_with(
                        """{} "param" with value "{}" is not of type "{}" or None.""".format(
                            cls.__name__, bad_data, expected_type
                        ),
                        luigi.parameter.OptionalParameterTypeWarning,
                    )

        # Test with value from config
        assert luigi.build([TestConfig()], local_scheduler=True)

    def test_optional_str_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalStrParameter` class."""
        with set_luigi_config({"TestConfig": {"param": "expected value", "empty_param": ""}}):
            self.actual_test(
                luigi.parameter.OptionalStrParameter,
                None,
                "expected value",
                "str",
                0,
            )
            self.actual_test(
                luigi.parameter.OptionalStrParameter,
                "default value",
                "expected value",
                "str",
                0,
            )

    def test_optional_int_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalIntParameter` class."""
        with set_luigi_config({"TestConfig": {"param": "10", "empty_param": ""}}):
            self.actual_test(luigi.parameter.OptionalIntParameter, None, 10, "int", "bad data")
            self.actual_test(luigi.parameter.OptionalIntParameter, 1, 10, "int", "bad data")

    def test_optional_bool_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalBoolParameter` class."""
        with set_luigi_config({"TestConfig": {"param": "true", "empty_param": ""}}):
            self.actual_test(
                luigi_tools.parameter.OptionalBoolParameter,
                None,
                True,
                "bool",
                "bad data",
            )
            self.actual_test(
                luigi_tools.parameter.OptionalBoolParameter,
                False,
                True,
                "bool",
                "bad data",
            )

    def test_optional_float_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalFloatParameter` class."""
        with set_luigi_config({"TestConfig": {"param": "10.5", "empty_param": ""}}):
            self.actual_test(
                luigi.parameter.OptionalFloatParameter,
                None,
                10.5,
                "float",
                "bad data",
            )
            self.actual_test(
                luigi.parameter.OptionalFloatParameter,
                1.5,
                10.5,
                "float",
                "bad data",
            )

    def test_optional_dict_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalDictParameter` class."""
        with set_luigi_config({"TestConfig": {"param": '{"a": 10}', "empty_param": ""}}):
            self.actual_test(
                luigi.parameter.OptionalDictParameter,
                None,
                {"a": 10},
                "FrozenOrderedDict",
                "bad data",
            )
            self.actual_test(
                luigi.parameter.OptionalDictParameter,
                {"a": 1},
                {"a": 10},
                "FrozenOrderedDict",
                "bad data",
            )

    def test_optional_list_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalListParameter` class."""
        with set_luigi_config({"TestConfig": {"param": "[10.5]", "empty_param": ""}}):
            self.actual_test(
                luigi.parameter.OptionalListParameter,
                None,
                (10.5,),
                "tuple",
                "bad data",
            )
            self.actual_test(
                luigi.parameter.OptionalListParameter,
                (1.5,),
                (10.5,),
                "tuple",
                "bad data",
            )

    def test_optional_tuple_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalTupleParameter` class."""
        with set_luigi_config({"TestConfig": {"param": "[10.5]", "empty_param": ""}}):
            self.actual_test(
                luigi.parameter.OptionalTupleParameter,
                None,
                (10.5,),
                "tuple",
                "bad data",
            )
            self.actual_test(
                luigi.parameter.OptionalTupleParameter,
                (1.5,),
                (10.5,),
                "tuple",
                "bad data",
            )

    def test_optional_numerical_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalNumericalParameter` class."""
        with set_luigi_config({"TestConfig": {"param": "10.5", "empty_param": ""}}):
            self.actual_test(
                luigi.parameter.OptionalNumericalParameter,
                None,
                10.5,
                "float",
                "bad data",
                var_type=float,
                min_value=0,
                max_value=100,
            )
            self.actual_test(
                luigi.parameter.OptionalNumericalParameter,
                1.5,
                10.5,
                "float",
                "bad data",
                var_type=float,
                min_value=0,
                max_value=100,
            )

    def test_optional_choice_parameter(self, tmp_working_dir):
        """Test the `luigi.parameter.OptionalChoiceParameter` class."""
        with set_luigi_config({"TestConfig": {"param": "expected value", "empty_param": ""}}):
            choices = ["default value", "expected value", "null"]
            self.actual_test(
                luigi.parameter.OptionalChoiceParameter,
                None,
                "expected value",
                "str",
                "bad data",
                choices=choices,
            )
            self.actual_test(
                luigi.parameter.OptionalChoiceParameter,
                "default value",
                "expected value",
                "str",
                "bad data",
                choices=choices,
            )


class TestBoolParameter:
    """Test the `luigi_tools.parameter.BoolParameter` class."""

    def test_auto_set_parsing(self):
        """Test that parsing is properly fixed according to the default value."""

        class TaskBoolParameter(luigi.Task):
            """A simple test task."""

            a = luigi_tools.parameter.BoolParameter(default=True)
            b = luigi_tools.parameter.BoolParameter(default=False)
            c = luigi_tools.parameter.BoolParameter(default=False, parsing="explicit")
            d = luigi_tools.parameter.BoolParameter()

            def run(self):
                assert self.a is True
                assert self.b is False
                assert self.c is False
                assert self.d is False and self.d is not None

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        assert TaskBoolParameter.a.parsing == "explicit"
        assert TaskBoolParameter.b.parsing == "implicit"
        assert TaskBoolParameter.c.parsing == "explicit"

        with set_luigi_config():
            assert luigi.build([TaskBoolParameter()], local_scheduler=True)

    def test_true_implicit_failing(self):
        """Test that setting `default=True` and `parsing="implicit"` raises an exception."""
        # pylint: disable=unused-variable
        with pytest.raises(ValueError):

            class TaskBoolParameterFail(luigi.Task):
                """A simple test task."""

                a = luigi_tools.parameter.BoolParameter(default=True, parsing="implicit")


@pytest.mark.parametrize("default", [None, "not_existing_dir"])
@pytest.mark.parametrize("absolute", [True, False])
@pytest.mark.parametrize("exists", [True, False])
def test_path_parameter(tmpdir, default, absolute, exists):
    """Test the `luigi.parameter.PathParameter` class."""

    class TaskPathParameter(luigi.Task):
        """A simple test task."""

        a = luigi.parameter.PathParameter(
            default=str(tmpdir / default) if default is not None else str(tmpdir),
            absolute=absolute,
            exists=exists,
        )
        b = luigi.parameter.OptionalPathParameter(
            default=str(tmpdir / default) if default is not None else str(tmpdir),
            absolute=absolute,
            exists=exists,
        )
        c = luigi.parameter.OptionalPathParameter(default=None)
        d = luigi.parameter.OptionalPathParameter(default="not empty default")

        def run(self):
            # Use the parameter as a Path object
            new_file = self.a / "test.file"
            new_optional_file = self.b / "test_optional.file"
            if default is not None:
                new_file.parent.mkdir(parents=True)
            new_file.touch()
            new_optional_file.touch()
            assert new_file.exists()
            assert new_optional_file.exists()
            assert self.c is None
            assert self.d is None

        def output(self):
            return luigi.LocalTarget("not_existing_file")

    # Test with default values
    with set_luigi_config({"TaskPathParameter": {"d": ""}}):
        if default is not None and exists:
            with pytest.raises(ValueError, match="The path .* does not exist"):
                luigi.build([TaskPathParameter()], local_scheduler=True)
        else:
            assert luigi.build([TaskPathParameter()], local_scheduler=True)

    # Test with values from config
    with set_luigi_config(
        {
            "TaskPathParameter": {
                "a": str(tmpdir / (default + "_from_config"))
                if default is not None
                else str(tmpdir),
                "b": str(tmpdir / (default + "_from_config"))
                if default is not None
                else str(tmpdir),
                "d": "",
            }
        }
    ):
        if default is not None and exists:
            with pytest.raises(ValueError, match="The path .* does not exist"):
                luigi.build([TaskPathParameter()], local_scheduler=True)
        else:
            assert luigi.build([TaskPathParameter()], local_scheduler=True)


def _check_parameter(*, obj, parameter, expected_dict, override_comparison_fields=None):
    """Check DataclassParameter for simple objects."""
    string = parameter.serialize(obj)
    assert string == json.dumps(expected_dict)

    dictionary = parameter.parse(string)
    assert dictionary == expected_dict

    new_obj1 = parameter.normalize(dictionary)
    new_obj2 = parameter.normalize(obj)

    for new_obj in (new_obj1, new_obj2):
        # manually compare the attributes for expected differences with origin obj
        if override_comparison_fields:
            for attribute_name, attribute_value in override_comparison_fields.items():
                actual_value = getattr(new_obj, attribute_name)
                assert actual_value == attribute_value
        else:
            # compare recursively
            _compare(new_obj, obj)


def _compare(actual, expected):
    """Recursively compare the two objects."""
    if dataclasses.is_dataclass(expected):
        assert type(actual) is type(expected), (
            f"Actual Class {type(actual)}. " f"Expected Class {type(expected)}."
        )
        for f in dataclasses.fields(expected):
            actual_value = getattr(actual, f.name)
            expected_value = getattr(expected, f.name)
            _compare(actual_value, expected_value)
    elif isinstance(expected, collections.abc.Mapping):
        assert isinstance(actual, FrozenOrderedDict)
        for key in expected:
            _compare(actual[key], expected[key])
    elif isinstance(expected, collections.abc.Sequence) and not isinstance(expected, str):
        assert isinstance(actual, tuple)
        for i, value in enumerate(expected):
            _compare(actual[i], value)
    else:
        assert actual == expected


def test_DataclassParameter__raises_no_dataclass():
    """Test guard against passing non dataclasses."""

    class A:
        a: int

    with pytest.raises(TypeError, match=r"Class type \'A\' is not a dataclass."):
        luigi_tools.parameter.DataclassParameter(cls_type=A)


def test_DataclassParameter__primitives():
    """Test the DataclassParameter with primitive types."""

    @dataclasses.dataclass
    class MyClass:
        a: str
        b: int
        c: float
        d: bool

    parameter = luigi_tools.parameter.DataclassParameter(cls_type=MyClass)

    data = {"a": "1", "b": 2, "c": 3.0, "d": False}

    instance = MyClass(**data)

    _check_parameter(obj=instance, parameter=parameter, expected_dict=data)


def test_DataclassParameter__primitives__ordering():
    """Test the DataclassParameter with differing primitive ordering."""

    @dataclasses.dataclass(frozen=True, eq=True)
    class A:
        a: str
        b: int
        c: float
        d: bool

    @dataclasses.dataclass(frozen=True, eq=True)
    class B:
        b: int
        a: str
        d: bool
        c: float

    a = A(a="1", b=2, c=3.0, d=False)
    b = B(a="1", b=2, c=3.0, d=False)

    pA = luigi_tools.parameter.DataclassParameter(cls_type=A)
    pB = luigi_tools.parameter.DataclassParameter(cls_type=B)

    # ordering affects serialization
    sA = pA.serialize(a)
    sB = pB.serialize(b)
    assert sA != sB

    # but not dictionary comparison
    dA = pA.parse(sA)
    dB = pB.parse(sB)
    assert dA == dB

    # ordering also affects deserialization and hashes
    nA = pA.normalize(a)
    nB = pB.normalize(b)
    assert nA != nB
    assert hash(nA) != hash(nB)

    # ordering also affects deserialization and hashes
    nA = pA.normalize(dA)
    nB = pB.normalize(dB)
    assert nA != nB
    assert hash(nA) != hash(nB)


def test_DataclassParameter__sequences():
    """Test the DataclassParameter with sequence types."""

    @dataclasses.dataclass(frozen=True, eq=True)
    class A:
        a: list
        b: tuple
        c: typing.List[str]
        d: typing.List[int]
        e: typing.Tuple[float]

    a = A(a=[1, "1"], b=(2.0, "2"), c=["a", "b"], d=[3, 4], e=(5.0, 6.0))

    p = luigi_tools.parameter.DataclassParameter(cls_type=A)

    # Note how the e tuple has been converted to a list with json serialization
    expected_dict = {"a": [1, "1"], "b": [2.0, "2"], "c": ["a", "b"], "d": [3, 4], "e": [5.0, 6.0]}

    _check_parameter(obj=a, parameter=p, expected_dict=expected_dict)


def test_DataclassParameter__mappings():
    """Test the DataclassParameter with mapping types."""

    @dataclasses.dataclass(frozen=True, eq=True)
    class A:
        a: dict
        b: collections.OrderedDict
        c: typing.Dict[str, int]
        d: typing.Dict[int, float]

    a = A(
        a={"a": 1, "b": "2"},
        b=collections.OrderedDict([("c", 3.0), ("d", 4)]),
        c={"e": 5, "f": 6},
        d={7: 8.0, 9: 10.0},
    )

    p = luigi_tools.parameter.DataclassParameter(cls_type=A)

    # Note how json serialization converts d keys into strings
    expected_dict = {
        "a": {"a": 1, "b": "2"},
        "b": {"c": 3.0, "d": 4},
        "c": {"e": 5, "f": 6},
        "d": {"7": 8.0, "9": 10.0},
    }
    _check_parameter(obj=a, parameter=p, expected_dict=expected_dict)


def test_DataclassParameter__nesting():
    """Test the DataclassParameter with nested dataclass objects."""

    @dataclasses.dataclass(frozen=True, eq=True)
    class A:
        a: str
        b: float

    @dataclasses.dataclass(frozen=True, eq=True)
    class B:
        c: A
        d: float

    @dataclasses.dataclass(frozen=True, eq=True)
    class C:
        e: A
        f: B

    c = C(e=A(a="1", b=2.0), f=B(c=A(a="2", b=3.0), d=4.0))

    p = luigi_tools.parameter.DataclassParameter(cls_type=C)

    expected_dict = {"e": {"a": "1", "b": 2.0}, "f": {"c": {"a": "2", "b": 3.0}, "d": 4.0}}

    _check_parameter(obj=c, parameter=p, expected_dict=expected_dict)


def test_DataclassParameter__nesting__2():
    """Test the DataclassParameter with nested dataclass objects."""

    @dataclasses.dataclass(frozen=True, eq=True)
    class A:
        a: str
        b: float

    @dataclasses.dataclass(frozen=True, eq=True)
    class B:
        c: A
        d: float

    @dataclasses.dataclass(frozen=True, eq=True)
    class C:
        e: typing.List[A]
        f: typing.Dict[str, B]

    p = luigi_tools.parameter.DataclassParameter(cls_type=C)

    obj = C(
        e=[A(a="1", b=2.0), A(a="2", b=3.0)],
        f={
            "4": B(c=A(a="5", b=6.0), d=7.0),
            "5": B(c=A(a="8", b=9.0), d=10.0),
        },
    )
    _check_parameter(
        obj=obj,
        parameter=p,
        expected_dict={
            "e": [
                {"a": "1", "b": 2.0},
                {"a": "2", "b": 3.0},
            ],
            "f": {
                "4": {"c": {"a": "5", "b": 6.0}, "d": 7.0},
                "5": {"c": {"a": "8", "b": 9.0}, "d": 10.0},
            },
        },
    )


def test_DataclassParameter__Optional__Union():
    """Test the DataclassParameter with Optional attributes."""

    @dataclasses.dataclass(frozen=True, eq=True)
    class A:
        a: int
        b: typing.Union[int, str]
        c: typing.Optional[float] = None
        d: typing.Optional[typing.List[int]] = None
        e: typing.Optional[typing.Dict[str, float]] = None

    a1 = A(a=1, b=2, c=3.0, d=[4, 5], e={"6": 7.0, "8": 9.0})
    a2 = A(a=1, b="2")

    p = luigi_tools.parameter.DataclassParameter(cls_type=A)

    _check_parameter(
        obj=a1,
        parameter=p,
        expected_dict={"a": 1, "b": 2, "c": 3.0, "d": [4, 5], "e": {"6": 7.0, "8": 9.0}},
    )
    _check_parameter(
        obj=a2,
        parameter=p,
        expected_dict={"a": 1, "b": "2", "c": None, "d": None, "e": None},
    )


def test_DataclassParameter__Optional_dataclasses():
    """Test DataclassParameter with optional dataclass attributes."""

    @dataclasses.dataclass
    class A:
        a: int

    @dataclasses.dataclass
    class B:
        b: typing.Optional[A] = None

    obj = B(b=A(a=1))

    p = luigi_tools.parameter.DataclassParameter(cls_type=B)

    expected_dict = {"b": {"a": 1}}

    string = p.serialize(obj)
    assert string == json.dumps(expected_dict)

    dictionary = p.parse(string)
    assert dictionary == expected_dict

    obj = p.normalize(dictionary)
    assert isinstance(obj.b, A), type(obj.b)

    obj = B()

    p = luigi_tools.parameter.DataclassParameter(cls_type=B)

    expected_dict = {"b": None}

    string = p.serialize(obj)
    assert string == json.dumps(expected_dict)

    dictionary = p.parse(string)
    assert dictionary == expected_dict

    obj = p.normalize(dictionary)
    assert obj.b is None


def test_DataclassParameter__Any():
    """Test the DataclassParameter with typing.Any attributes."""

    @dataclasses.dataclass(frozen=True, eq=True)
    class D:
        a: int

    @dataclasses.dataclass(frozen=True, eq=True)
    class A:
        a: typing.Any
        b: typing.Any
        c: typing.List[typing.Any]
        d: typing.Dict[str, typing.Any]
        e: typing.Optional[typing.Dict[int, typing.Any]]

    a = A(
        a="1",
        b=["1", 2, 3.0, [D(a=4)]],
        c=[5, "6", 7.0, [D(a=8)]],
        d={"9": 10.0, "11": "12"},
        e={13: 14, 15: "16", 17: {"a": D(a=18)}},
    )

    p = luigi_tools.parameter.DataclassParameter(cls_type=A)

    expected_dict = {
        "a": "1",
        "b": ["1", 2, 3.0, [{"a": 4}]],
        "c": [5, "6", 7.0, [{"a": 8}]],
        "d": {"9": 10.0, "11": "12"},
        "e": {"13": 14, "15": "16", "17": {"a": {"a": 18}}},
    }

    string = p.serialize(a)
    assert string == json.dumps(expected_dict)

    dictionary = p.parse(string)
    assert dictionary == expected_dict

    obj = p.normalize(dictionary)
    assert obj.a == "1"

    # Note: It is not possible to reconstruct the D dataclass if the type is Any
    assert obj.b == ("1", 2, 3.0, (FrozenOrderedDict([("a", 4)]),))
    assert obj.c == (5, "6", 7.0, (FrozenOrderedDict([("a", 8)]),))
    assert obj.d == FrozenOrderedDict(a.d)
    assert obj.e == FrozenOrderedDict(
        [
            (13, 14),
            (15, "16"),
            (17, FrozenOrderedDict([("a", FrozenOrderedDict([("a", 18)]))])),
        ]
    )


def test_DataclassParameter__build():
    """Test DataclassParameter using a config and luigi build."""

    @dataclasses.dataclass(frozen=True, eq=True)
    class ParamA:
        a: dict
        b: collections.OrderedDict
        c: typing.Dict[str, int]
        d: typing.Dict[int, float]

    class TaskDataClassParameter(luigi.Task):
        """A simple test task."""

        a = luigi_tools.parameter.DataclassParameter(cls_type=ParamA)

        def run(self):
            # pylint: disable=no-member
            assert self.a.a == {"a": 1, "b": "2"}
            assert self.a.b == {"c": 3.0, "d": 4}
            assert self.a.c == {"e": 5, "f": 6}
            assert self.a.d == {7: 8.0, 9: 10.0}

        def output(self):
            return luigi.LocalTarget("not_existing_file")

    # Test with values from config
    with set_luigi_config(
        {
            "TaskDataClassParameter": {
                "a": """{
                    "a": {"a": 1, "b": "2"},
                    "b": {"c": 3.0, "d": 4},
                    "c": {"e": 5, "f": 6},
                    "d": {"7": 8.0, "9": 10.0}
                }"""
            }
        }
    ):
        assert luigi.build([TaskDataClassParameter()], local_scheduler=True)

    # Parsing fails with integer keys
    with set_luigi_config(
        {
            "TaskDataClassParameter": {
                "a": """{
                    "a": {"a": 1, "b": "2"},
                    "b": {"c": 3.0, "d": 4},
                    "c": {"e": 5, "f": 6},
                    "d": {7: 8.0, 9: 10.0}
                }"""
            }
        }
    ):
        with pytest.raises(ValueError):
            luigi.build([TaskDataClassParameter()], local_scheduler=True)

    # But the constructor works with either integers and strings as keys
    assert luigi.build(
        [
            TaskDataClassParameter(
                {
                    "a": {"a": 1, "b": "2"},
                    "b": {"c": 3.0, "d": 4},
                    "c": {"e": 5, "f": 6},
                    "d": {7: 8.0, 9: 10.0},
                }
            )
        ],
        local_scheduler=True,
    )

    assert luigi.build(
        [
            TaskDataClassParameter(
                {
                    "a": {"a": 1, "b": "2"},
                    "b": {"c": 3.0, "d": 4},
                    "c": {"e": 5, "f": 6},
                    "d": {"7": 8.0, "9": 10.0},
                }
            )
        ],
        local_scheduler=True,
    )



# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The luigi-tools package.

Extend and add new features to the luigi package.
"""
import warnings

from pkg_resources import get_distribution

__version__ = get_distribution("luigi-tools").version


class MovedToLuigiWarning(DeprecationWarning):
    """Warning for features moved to the official luigi package."""


def moved_to_luigi_warning(
    luigi_version=None, previous_luigi_version=None, deprecation_version=None
):
    """Raise a deprecation warning for features moved to official luigi package.

    Args:
        luigi_version (str): The version of the luigi package from which the feature is available.
        previous_luigi_version (str): The version of the ``luigi`` package after which the feature
            is available (use this when the feature is merged into the master branch of the
            ``luigi`` package but it is not released yet).
        deprecation_version (str): The version of the ``luigi_tools`` package from which the feature
            will be removed.
    """
    if (luigi_version is None) == (previous_luigi_version is None):
        raise ValueError(
            "Either the 'luigi_version' or the 'previous_luigi_version' argument must be not None "
            "but not both of them"
        )
    if luigi_version is not None:
        msg = (
            "This feature was moved to the luigi package and is available since version "
            f"{luigi_version}."
        )
    else:
        msg = (
            "This feature was moved to the luigi package and will be available after version "
            f"{previous_luigi_version}."
        )

    if deprecation_version is not None:
        msg += f" It will be deprecated in version {deprecation_version}."
    warnings.warn(
        msg,
        MovedToLuigiWarning,
    )



# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides some functions to work with luigi tasks."""
import configparser
import logging
import os
import re
import warnings
from configparser import ConfigParser
from pathlib import Path

import luigi
from luigi.parameter import _no_value as PARAM_NO_VALUE

L = logging.getLogger(__name__)


class WorkflowError(Exception):
    """Exception raised when the workflow is not consistent."""


def recursive_check(task, attr="rerun"):
    """Check if a task or any of its recursive dependencies has a given attribute set to True."""
    val = getattr(task, attr, False)

    for dep in task.deps():
        val = val or getattr(dep, attr, False) or recursive_check(dep, attr)

    return val


def target_remove(target, *args, **kwargs):
    """Remove a given target by calling its 'exists()' and 'remove()' methods."""
    try:
        if target.exists():
            try:
                L.debug("Removing %s", target.path)
            except AttributeError:  # pragma: no cover
                pass
            target.remove()
    except AttributeError as e:
        raise AttributeError("The target must have 'exists()' and 'remove()' methods") from e


def apply_over_luigi_iterable(luigi_iterable, func):
    """Apply the given function to a luigi iterable (task.input() or task.output())."""
    try:
        results = {}
        for key, i in luigi_iterable.items():
            results[key] = func(i, key)
    except AttributeError:
        results = []
        for i in luigi.task.flatten(luigi_iterable):
            results.append(func(i))
    return results


def apply_over_inputs(task, func):
    """Apply the given function to all inputs of a :class:`luigi.Task`.

    The given function should accept the following arguments:
        * task_output: the output(s) of the required task(s) of the given task
        * key=None: the key when the iterable is a dictionary
    """
    inputs = task.input()
    return apply_over_luigi_iterable(inputs, func)


def apply_over_outputs(task, func):
    """Apply the given function to all outputs of a :class:`luigi.Task`.

    The given function should accept the following arguments:
        * target: the output target(s) of the given task
        * key=None: the key when the iterable is a dictionary
    """
    outputs = task.output()

    return apply_over_luigi_iterable(outputs, func)


def apply_over_required(task, func):
    """Apply the given function to all required tasks of a :class:`luigi.WrapperTask`.

    The given function should accept the following arguments:
        * target: the output target(s) of the given task
        * key=None: the key when the iterable is a dictionary
    """
    required = task.requires()

    return apply_over_luigi_iterable(required, func)


def get_dependency_graph(
    task,
    allow_orphans=False,
):
    """Compute dependency graph of a given task.

    Args:
        task (luigi.Task): the task from which the dependency graph is computed.
        allow_orphans (bool): If set to True, orphan nodes are returned with a None child.

    Returns:
        list(luigi.Task): A list of (parent, child) tuples.
    """
    children = []
    for t in task.deps():
        children.append((task, t))
        children.extend(get_dependency_graph(t))
    if not children and allow_orphans:
        children.append((task, None))
    return children


def _param_repr(description, default):
    """Produce parameter help string representation for sphinx-doc."""
    if description:
        help_str = description
    else:
        help_str = ""
    if default is not PARAM_NO_VALUE:
        help_str += f"({default})"
    return help_str


def graphviz_dependency_graph(
    g,
    graph_attrs=None,
    node_attrs=None,
    edge_attrs=None,
    root_attrs=None,
    task_names=None,
    node_kwargs=None,
    edge_kwargs=None,
    graphviz_class=None,
):
    """Create a GraphViz dependency graph.

    Args:
        g (list(tuple(luigi.Task))): A list of tuples of :class:`luigi.Task` objects, usually
            created with :mod:`luigi_tools.util.get_dependency_graph()`.
        graph_attrs (dict): The graph attributes that will override the defaults.
        node_attrs (dict): The node attributes that will override the defaults.
        edge_attrs (dict): The edge attributes that will override the defaults.
        root_attrs (dict): The specific node attributes only used for the root that will override
            the defaults.
        task_names (dict): A dictionary with :class:`luigi.Task` objects as keys and custom names as
            values.
        node_kwargs (dict): A dictionary with :class:`luigi.Task` objects as keys and the kwargs
            given to :meth:`graphviz.Digraph.node()` as values.
        edge_kwargs (dict): A dictionary with :class:`luigi.Task` objects as keys and the kwargs
            given to :meth:`graphviz.Digraph.edge()` as values.
        graphviz_class (graphviz.Graph or graphviz.Digraph): The class used to store the graph.

    Returns:
        The Graph or Digraph created.
    """
    if len(g) <= 0:
        raise ValueError("The dependency graph is empty")

    # Setup default values if they are not given
    if graphviz_class is None:
        try:
            from graphviz import Digraph  # pylint: disable=import-outside-toplevel

            graphviz_class = Digraph
        except ImportError as e:  # pragma: no cover
            msg = "Could not import GraphViz, please install it."
            raise ImportError(msg) from e

    default_graph_attrs = {
        "size": "7.0, 15.0",
        "bgcolor": "transparent",
        "fontsize": "9",
        "layout": "dot",
        "rankdir": "TB",
        "dpi": "300",
    }
    if graph_attrs is not None:
        default_graph_attrs.update(graph_attrs)

    default_node_attrs = {
        "shape": "box",
        "fontsize": "9",
        "height": "0.25",
        "fontname": '"Vera Sans, DejaVu Sans, Liberation Sans, Arial, Helvetica, sans"',
        "style": "setlinewidth(0.5),filled",
        "fillcolor": "white",
    }
    if node_attrs is not None:
        default_node_attrs.update(node_attrs)

    default_edge_attrs = {
        "arrowsize": "0.5",
        "style": "setlinewidth(0.5)",
    }
    if edge_attrs is not None:
        default_edge_attrs.update(edge_attrs)

    default_root_attrs = {"color": "red", "penwidth": "1.5"}
    if root_attrs is not None:
        default_root_attrs.update(root_attrs)

    if not task_names:
        task_names = {}

    if not node_kwargs:
        node_kwargs = {}

    if not edge_kwargs:
        edge_kwargs = {}

    # Build the dependency graph
    dot = graphviz_class(
        comment="Dependency graph",
        strict=True,
        graph_attr=default_graph_attrs,
        node_attr=default_node_attrs,
        edge_attr=default_edge_attrs,
    )

    root = g[0][0]
    task_name = task_names.get(root, root.__class__.__name__)
    dot.node(task_name, **default_root_attrs)

    for parent, child in g:
        if child is None:
            continue
        parent_name = task_names.get(parent, parent.__class__.__name__)
        child_name = task_names.get(child, child.__class__.__name__)
        dot.node(child_name, **node_kwargs.get(child, {}))
        dot.edge(parent_name, child_name, **edge_kwargs.get((parent, child), {}))

    return dot


def render_dependency_graph(graph, filepath, **kwargs):
    """Render a dependency graph using GraphViz.

    Args:
        graph (graphviz.Graph or graphviz.Digraph): The graph to render, usually create with
            :mod:`luigi_tools.util.graphviz_dependency_graph()`.
        filepath (str): The path to the rendered file. If the ``format`` kwarg is not given, this
            path should contain an extension. Otherwise, the extension is guessed from the format.
        kwargs: The kwargs are passed to the :meth:`graphviz.Digraph.render()` function.
    """
    filepath = Path(filepath)
    filename = filepath.with_suffix("")
    fileformat = kwargs.pop("format", None)
    if not fileformat:
        pattern = re.compile(r"\.?(.*)")
        match = re.match(pattern, filepath.suffix)
        fileformat = match.group(1) or None
    cleanup = kwargs.pop("cleanup", True)
    graph.render(filename=filename, format=fileformat, cleanup=cleanup, **kwargs)


def export_dependency_graph(
    task,
    filepath="dependency_graph.png",
    allow_orphans=False,
    graph_attrs=None,
    node_attrs=None,
    edge_attrs=None,
    root_attrs=None,
    task_names=None,
    node_kwargs=None,
    edge_kwargs=None,
    graphviz_class=None,
    **render_kwargs,
):
    """Build and export a dependency graph.

    Args:
        task (luigi.Task): the task from which the dependency graph is computed.
        filepath (str): The path to the rendered file. If the ``format`` kwarg is not given, this
            path should contain an extension. Otherwise, the extension is guessed from the format.
        allow_orphans (bool): If set to True, orphan nodes are returned with a None child.
        graph_attrs (dict): The graph attributes that will override the defaults.
        node_attrs (dict): The node attributes that will override the defaults.
        edge_attrs (dict): The edge attributes that will override the defaults.
        root_attrs (dict): The specific node attributes only used for the root that will override
            the defaults.
        task_names (dict): A dictionary with :class:`luigi.Task` objects as keys and custom names as
            values.
        node_kwargs (dict): A dictionary with :class:`luigi.Task` objects as keys and the kwargs
            given to :meth:`graphviz.Digraph.node()` as values.
        edge_kwargs (dict): A dictionary with :class:`luigi.Task` objects as keys and the kwargs
            given to :meth:`graphviz.Digraph.edge()` as values.
        graphviz_class (graphviz.Graph or graphviz.Digraph): The class used to store the graph.

    Keyword Args:
        render_kwargs: The keyword arguments are passed to the :meth:`graphviz.Digraph.render()`
            function.
    """
    # pylint: disable=too-many-arguments
    deps = get_dependency_graph(task, allow_orphans=allow_orphans)
    graph = graphviz_dependency_graph(
        deps,
        graph_attrs=graph_attrs,
        node_attrs=node_attrs,
        edge_attrs=edge_attrs,
        root_attrs=root_attrs,
        task_names=task_names,
        node_kwargs=node_kwargs,
        edge_kwargs=edge_kwargs,
        graphviz_class=graphviz_class,
    )
    render_dependency_graph(graph, filepath, **render_kwargs)


def register_templates(directory=None, name=None, hierarchy_end=True):
    """Add INI templates to the config file list processed by luigi.

    This function should be used before the :class:`luigi.Task` are processed, so it should usually
    be used in the ``__init__.py`` of your package.

    In order to use a template, a ``Template`` section entry must be added to the luigi.cfg of the
    project. This section should contain at least one entry with the name of the template to use,
    and, optionally, another entry with the path to the directory containing the templates, as the
    following example:

    .. code-block:: INI

        [Template]
        name = template_name
        directory = /path/to/the/template/directory

    If the ``directory`` or the ``name`` entries are given in the ``luigi.cfg`` file, they override
    the arguments given to this function.

    A template file should be similar to a ``luigi.cfg`` file with only the entries for which
    specific default values should be defined. The file should named according to the template
    name. For example, the template named ``template_1`` should be a INI file located in the given
    directory and named ``template_1.cfg``.

    Args:
        directory (str): Path to the directory containing the template files.
        name (str): The name of the template to use.
        hierarchy_end (bool): If set to False, the ``luigi.cfg`` entry is not added after the
            template entry, so it is possible to combine several templates before processing
            the ``luigi.cfg`` file.
    """
    base_config = luigi.configuration.get_config()

    if "Template" in base_config:
        template_config = base_config["Template"]
    else:
        template_config = {}

    template_path = template_config.get("directory", directory)
    if template_path is None:
        raise ValueError(
            "A directory must either be given to this function or in the [Template] section of "
            "the luigi.cfg file."
        )
    template_path = Path(template_path)

    template_name = template_config.get("name", name)
    if template_name is None:
        raise ValueError(
            "A name must either be given to this function or in the [Template] section of "
            "the luigi.cfg file."
        )

    template = (template_path / template_name).with_suffix(".cfg")
    if not template.exists():
        raise ValueError(f"The template '{template}' could not be found.")

    luigi.configuration.add_config_path(template)
    if hierarchy_end:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Config file does not exist.*")
            luigi.configuration.add_config_path("luigi.cfg")
            env_cfg = os.environ.get("LUIGI_CONFIG_PATH")
            if env_cfg is not None:
                luigi.configuration.add_config_path(env_cfg)


class set_luigi_config:
    """Context manager to set current luigi config.

    Args:
        params (dict): The parameters to load into the luigi configuration.
        configfile (str): Path to the temporary config file (luigi.cfg by default).
    """

    def __init__(self, params=None, configfile=None):
        self.params = params
        self.luigi_config = luigi.configuration.get_config()
        if configfile is not None:
            self.configfile = configfile
        else:
            self.configfile = "luigi.cfg"

    def __enter__(self):
        """Load the given luigi configuration."""
        # Reset luigi config
        self.luigi_config.clear()

        # Remove config file
        if os.path.exists(self.configfile):
            os.remove(self.configfile)

        # Export config
        if self.params is not None:
            self.export_config(self.configfile)

            # Set current config in luigi
            self.luigi_config.read(self.configfile)
        else:
            self.configfile = None

    def get_config(self):
        """Convert the parameter dict to a :class:`configparser.ConfigParser` object."""
        config = ConfigParser()
        config.read_dict(self.params)
        return config

    def export_config(self, filepath):
        """Export the configuration to the configuration file."""
        params = self.get_config()

        # Export params
        with open(filepath, "w") as configfile:  # pylint: disable=unspecified-encoding
            params.write(configfile)

    def __exit__(self, *args):
        """Reset the configuration when exiting the context manager."""
        # Remove config file
        if self.configfile is not None and os.path.exists(self.configfile):
            os.remove(self.configfile)

        # Reset luigi config
        self.luigi_config.clear()


def configparser_to_dict(luigi_config):
    """Transform a ConfigParser object to a dict."""
    dict_config = {}
    for section in luigi_config.sections():
        dict_config[section] = {}
        for option in luigi_config.options(section):
            dict_config[section][option] = luigi_config.get(section, option)
    return dict_config


def luigi_config_to_dict(filename="luigi.cfg"):
    """Load a luigi config file to a dict."""
    luigi_config = configparser.ConfigParser()
    luigi_config.read(filename)
    return configparser_to_dict(luigi_config)



# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for luigi-tools tasks."""

# pylint: disable=no-member
# pylint: disable=protected-access
# pylint: disable=too-many-lines
# pylint: disable=unused-argument
# pylint: disable=unused-variable
import json
import logging
import warnings

import luigi
import pytest
from luigi.util import inherits
from packaging.version import Version
from pkg_resources import get_distribution

import luigi_tools.parameter
import luigi_tools.target
import luigi_tools.task
import luigi_tools.util
from luigi_tools.task import DuplicatedParameterError
from luigi_tools.task import GlobalParameterNoValueError
from luigi_tools.util import set_luigi_config

from .tools import check_empty_file
from .tools import check_existing_file
from .tools import check_not_empty_file
from .tools import create_empty_file
from .tools import create_not_empty_file


class TestCopyParams:
    """Test the `luigi_tools.task.copy_params` decorator."""

    @pytest.mark.filterwarnings("ignore::UserWarning:luigi.parameter")
    def test_copy_params(self, tmpdir):
        """Simple test."""

        class TaskA(luigi.Task):
            """A simple test task."""

            a = luigi.Parameter(default="default_value_a")
            b = luigi.Parameter(default="default_value_b")

            def run(self):
                print(self.a)
                return self.a

            def output(self):
                return luigi.LocalTarget(tmpdir)

        @luigi_tools.task.copy_params(
            a=luigi_tools.task.ParamRef(TaskA),
            aa=luigi_tools.task.ParamRef(TaskA, "a"),
            a_default=luigi_tools.task.ParamRef(TaskA, "a", "given_default_value"),
            a_none=luigi_tools.task.ParamRef(TaskA, "a", None),
        )
        class TaskB(luigi.Task):
            """A simple test task."""

            b = luigi.Parameter(default="b")
            b_none = luigi.Parameter(default=None)

            def run(self):
                print(self.a, self.aa, self.a_default, self.a_none, self.b, self.b_none)
                return self.a, self.aa, self.a_default, self.a_none, self.b, self.b_none

            def output(self):
                return luigi.LocalTarget(tmpdir)

        # Test with default value
        task = TaskB()
        res = task.run()

        assert res == (
            "default_value_a",
            "default_value_a",
            "given_default_value",
            None,
            "b",
            None,
        )

        # Test with another value
        task = TaskB(a="new_a", aa="new_aa", a_default="new_default", b="bb")
        res = task.run()

        assert res == ("new_a", "new_aa", "new_default", None, "bb", None)

        # Empty copy_params arguments should raise a ValueError
        with pytest.raises(ValueError):

            @luigi_tools.task.copy_params()
            class TaskC(luigi.Task):
                """A simple test task."""

                a = luigi.Parameter(default="a")

        # Duplicated parameters should raise a DuplicatedParameterError
        with pytest.raises(DuplicatedParameterError):

            @luigi_tools.task.copy_params(
                a=luigi_tools.task.ParamRef(TaskA),
            )
            class TaskD(luigi.Task):
                """A simple test task."""

                a = luigi.Parameter(default="a")

        # Test with parameters that are serialized to generate the task ID
        class TaskWithListDictParams(luigi.Task):
            """A simple test task."""

            a = luigi.ListParameter(description="a in TaskWithListDictParams")
            b = luigi.DictParameter(description="b in TaskWithListDictParams")

            def run(self):
                assert self.a == (1, 2)
                assert self.b == {"attr1": 1, "attr2": 2}

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        @luigi_tools.task.copy_params(
            a_copy=luigi_tools.task.ParamRef(TaskWithListDictParams, "a"),
            b_copy=luigi_tools.task.ParamRef(TaskWithListDictParams, "b"),
        )
        class TaskCopyListDictParams(luigi.Task):
            """A simple test task."""

            a = luigi.ListParameter(description="a in TaskCopyListDictParams")
            b = luigi.DictParameter(description="b in TaskCopyListDictParams")

            def run(self):
                assert self.a == (1, 2)
                assert self.b == {"attr1": 1, "attr2": 2}
                assert self.a_copy == self.a
                assert self.b_copy == self.b

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with set_luigi_config(
            {
                "TaskWithListDictParams": {
                    "a": "[1, 2]",
                    "b": json.dumps({"attr1": 1, "attr2": 2}),
                },
                "TaskCopyListDictParams": {
                    "a": "[1, 2]",
                    "b": json.dumps({"attr1": 1, "attr2": 2}),
                    "a_copy": "[1, 2]",
                    "b_copy": json.dumps({"attr1": 1, "attr2": 2}),
                },
            }
        ):
            assert luigi.build([TaskCopyListDictParams()], local_scheduler=True)

    def type_test(
        self,
        cls_param,
        initial_value,
        str_initial_value,
        new_value,
        str_new_value,
        **kwargs,
    ):
        """Test function used for different types."""

        class TaskWithTypeParams(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A task with some parameters of given type."""

            a = cls_param(description="a in TaskWithTypeParams")

            def run(self):
                assert self.a == initial_value

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        @luigi_tools.task.copy_params(
            a_copy=luigi_tools.task.ParamRef(TaskWithTypeParams, "a"),
        )
        class TaskCopyTypeParams(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A task that copies some parameters of given type."""

            a = cls_param(description="a in TaskCopyTypeParams")
            with_value = luigi.BoolParameter()

            def run(self):
                if cls_param is luigi.parameter.OptionalListParameter:
                    assert self.a == tuple(initial_value)
                else:
                    assert self.a == initial_value
                if self.with_value:
                    assert self.a_copy == new_value
                else:
                    assert self.a_copy == self.a

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with set_luigi_config(
            {
                "TaskWithTypeParams": {
                    "a": str_initial_value,
                },
                "TaskCopyTypeParams": {
                    "a": str_initial_value,
                    "a_copy": str_new_value,
                },
            }
        ):
            assert luigi.build(
                [TaskWithTypeParams(), TaskCopyTypeParams(with_value=True)],
                local_scheduler=True,
            )

        with set_luigi_config(
            {
                "TaskWithTypeParams": {
                    "a": str_initial_value,
                },
                "TaskCopyTypeParams": {
                    "a": str_initial_value,
                },
            }
        ):
            assert luigi.build(
                [TaskWithTypeParams(), TaskCopyTypeParams(with_value=False)],
                local_scheduler=True,
            )
        print("test end with", cls_param)

    def test_int_param(self, tmp_working_dir):
        """Test with int parameters."""
        self.type_test(luigi.parameter.IntParameter, 1, "1", 2, "2")

    def test_float_param(self, tmp_working_dir):
        """Test with float parameters."""
        self.type_test(luigi.parameter.FloatParameter, 1.5, "1.5", 2.5, "2.5")

    def test_bool_param(self, tmp_working_dir):
        """Test with bool parameters."""
        self.type_test(luigi_tools.parameter.BoolParameter, True, "true", False, "false")

    def test_optional_list_param(self, tmp_working_dir):
        """Test with optional list parameters."""
        self.type_test(
            luigi.parameter.OptionalListParameter,
            (1, 2),
            "[1, 2]",
            (10, 20),
            "[10, 20]",
        )


class TestCopyParamsWithGlobals:
    """Test the `luigi_tools.task.copy_params` decorator with global parameters."""

    @pytest.mark.filterwarnings("ignore::UserWarning:luigi.parameter")
    @pytest.fixture
    def TaskA(self):
        """A simple test task."""

        class TaskA(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            a = luigi.Parameter(default="a")
            a_cfg = luigi.Parameter(default="a_cfg")

            def run(self):
                assert self.a == "a"
                assert self.a_cfg == "default_value_in_cfg"

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        return TaskA

    @pytest.mark.filterwarnings("ignore::UserWarning:luigi.parameter")
    def test_defaults(self, luigi_tools_working_directory, TaskA):
        """Test with default usage."""

        @luigi_tools.task.copy_params(
            aa=luigi_tools.task.ParamRef(TaskA, "a_cfg"),
            a_default=luigi_tools.task.ParamRef(TaskA, "a", "given_default_value"),
            a_none=luigi_tools.task.ParamRef(TaskA, "a", None),
        )
        class TaskB(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            b = luigi.Parameter(default="b")
            b_none = luigi.Parameter(default=None)
            mode = luigi.Parameter(default="default")

            def run(self):
                if self.mode == "default":
                    assert self.aa == "default_value_in_cfg"
                    assert self.a_default == "given_default_value"
                    assert self.b == "b"
                else:
                    assert self.aa == "constructor_value"
                    assert self.a_default == "new_default"
                    assert self.b == "bb"
                assert self.a_none is None
                assert self.b_none is None
                return self.aa, self.a_default, self.b

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        # Test with default value
        task = TaskB()
        res = task.run()

        assert res == ("default_value_in_cfg", "given_default_value", "b")

        # Test with another value
        task = TaskB(aa="constructor_value", a_default="new_default", b="bb", mode="constructor")
        res = task.run()

        assert res == ("constructor_value", "new_default", "bb")

        assert luigi.build([TaskA(), TaskB()], local_scheduler=True)
        assert luigi.build(
            [
                TaskA(),
                TaskB(
                    aa="constructor_value",
                    a_default="new_default",
                    b="bb",
                    mode="constructor",
                ),
            ],
            local_scheduler=True,
        )

    def test_empty(self):
        """Test with empty params.

        Empty copy_params arguments should raise a ValueError.
        """
        with pytest.raises(ValueError):

            @luigi_tools.task.copy_params()
            class TaskC(luigi.Task):
                """A simple test task."""

                a = luigi.Parameter(default="a")

    def test_duplicated(self, TaskA):
        """Test with duplicated params.

        Duplicated parameters should raise a DuplicatedParameterError.
        """
        with pytest.raises(DuplicatedParameterError):

            @luigi_tools.task.copy_params(
                a=luigi_tools.task.ParamRef(TaskA),
            )
            class TaskD(luigi_tools.task.GlobalParamMixin, luigi.Task):
                """A simple test task."""

                a = luigi.Parameter(default="a")

    @pytest.mark.filterwarnings("ignore::UserWarning:luigi.parameter")
    def test_no_default_value(self):
        """Test without default value.

        Global parameter with _no_default_value should raise GlobalParameterNoValueError.
        """

        class TaskE(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            e = luigi.Parameter(default=luigi_tools.task._no_default_value)

            def run(self):
                assert self.e == luigi_tools.task._no_default_value

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with pytest.raises(GlobalParameterNoValueError):
            assert luigi.build([TaskE()], local_scheduler=True)

    @pytest.mark.filterwarnings("ignore::UserWarning:luigi.parameter")
    def test_copy_params_no_default_value(self, TaskA):
        """Test copied param without default value.

        Global parameter with _no_default_value should raise GlobalParameterNoValueError.
        """

        @luigi_tools.task.copy_params(
            a=luigi_tools.task.ParamRef(TaskA),
        )
        class TaskF(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            f = luigi.Parameter(default=luigi_tools.task._no_default_value)

            def run(self):
                assert self.a == "a"
                assert self.f == luigi_tools.task._no_default_value

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with pytest.raises(GlobalParameterNoValueError):
            luigi.build([TaskF()], local_scheduler=True)

    def test_inherits(self):
        """Compare with luigi.util.inherits."""

        class TaskG(luigi.Task):
            """A simple test task."""

            g = luigi.Parameter(default="default_value_g")

            def run(self):
                assert self.g == "new_value"

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        @inherits(TaskG)
        class TaskH(luigi.Task):
            """A simple test task."""

            h = luigi.Parameter()

            def requires(self):
                return self.clone(TaskG)

            def run(self):
                assert self.g == "another_new_value"
                assert self.h == "h_from_cfg"

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with set_luigi_config(
            {
                "TaskG": {
                    "g": "new_value",
                },
                "TaskH": {
                    "g": "another_new_value",
                    "h": "h_from_cfg",
                },
            }
        ):
            failed_task = []
            exceptions = []

            @TaskG.event_handler(luigi.Event.FAILURE)
            def check_exception(task, exception):
                failed_task.append(str(task))
                exceptions.append(str(exception))

            # The workflow fails because TaskG().g == "another_new_value" != "new_value"
            assert not luigi.build([TaskH()], local_scheduler=True)
            assert failed_task == [str(TaskG(g="another_new_value"))]
            assert exceptions == [
                "assert 'another_new_value' == 'new_value'\n  - new_value\n  + another_new_value"
            ]

    def test_compare_inherits(self, luigi_tools_working_directory):
        """Compare with luigi.util.inherits."""

        class TaskI(luigi.Task):
            """A simple test task."""

            i = luigi.Parameter(default="default_value_i")

            def run(self):
                assert self.i == "new_value"
                create_not_empty_file(self.output().path)

            def output(self):
                return luigi.LocalTarget(luigi_tools_working_directory / "TaskI_output.test")

        @luigi_tools.task.copy_params(
            i=luigi_tools.task.ParamRef(TaskI),
        )
        class TaskJ(luigi.Task):
            """A simple test task."""

            j = luigi.Parameter()

            def requires(self):
                return TaskI()

            def run(self):
                assert self.i == "another_new_value"
                assert self.j == "j_from_cfg"

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with set_luigi_config(
            {
                "TaskI": {
                    "i": "new_value",
                },
                "TaskJ": {
                    "i": "another_new_value",
                    "j": "j_from_cfg",
                },
            }
        ):
            # The workflow now succeeds because TaskI().i == "new_value" and
            # TaskJ().i == "another_new_value"
            assert luigi.build([TaskJ()], local_scheduler=True)

    def test_serialized_parameters(self):
        """Test with parameters that are serialized to generate the task ID."""

        class GlobalParamTaskWithListDictParams(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            a = luigi.ListParameter(default="a in GlobalParamTaskWithListDictParams")
            b = luigi.DictParameter(default="b in GlobalParamTaskWithListDictParams")

            def run(self):
                assert self.a == (1, 2)
                assert self.b == {"attr1": 1, "attr2": 2}

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        @luigi_tools.task.copy_params(
            a_copy=luigi_tools.task.ParamRef(GlobalParamTaskWithListDictParams, "a"),
            b_copy=luigi_tools.task.ParamRef(GlobalParamTaskWithListDictParams, "b"),
        )
        class GlobalParamTaskCopyListDictParams(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            a_new = luigi.ListParameter(default="a in GlobalParamTaskCopyListDictParams")
            b_new = luigi.DictParameter(default="b in GlobalParamTaskCopyListDictParams")

            def run(self):
                assert self.a_new == (1, 2)
                assert self.b_new == {"attr1": 1, "attr2": 2}
                assert self.a_copy == self.a_new
                assert self.b_copy == self.b_new

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with set_luigi_config(
            {
                "GlobalParamTaskWithListDictParams": {
                    "a": "[1, 2]",
                    "b": json.dumps({"attr1": 1, "attr2": 2}),
                },
                "GlobalParamTaskCopyListDictParams": {
                    "a_new": "[1, 2]",
                    "b_new": json.dumps({"attr1": 1, "attr2": 2}),
                },
            }
        ):
            assert luigi.build(
                [
                    GlobalParamTaskWithListDictParams(),
                    GlobalParamTaskCopyListDictParams(),
                ],
                local_scheduler=True,
            )

    def test_not_comparable_attribute(self):
        """Test with params that can't be compared."""

        class NotComparable:
            """A class that always raises an exception when __eq__ is called."""

            def __eq__(self, other):
                raise TypeError(f"Can't compare {self.__class__} with {type(other)}")

        class GlobalParamTaskSetGetAttr(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            a = luigi.ListParameter(default="a in GlobalParamTaskSetGetAttr")
            b = None
            c = 1

            def run(self):
                assert self.a == "a in GlobalParamTaskSetGetAttr"
                self.a = NotComparable()
                assert isinstance(self.a, NotComparable)
                self.a = "a in GlobalParamTaskSetGetAttr"
                assert self.a == "a in GlobalParamTaskSetGetAttr"

                assert self.b is None
                self.b = NotComparable()
                assert isinstance(self.b, NotComparable)
                self.b = None
                assert self.b is None

                assert self.c == 1
                self.c = NotComparable()
                assert isinstance(self.c, NotComparable)
                self.c = 1
                assert self.c == 1

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        assert luigi.build(
            [GlobalParamTaskSetGetAttr()],
            local_scheduler=True,
        )

    def test_task_value(self):
        """Test monkey patch of task_value().

        This test fails without monkey patch because a list parameter can not be None.
        """

        class TaskListParameter(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            a = luigi.parameter.ListParameter(default=(1, 2))
            b = luigi.parameter.ListParameter()

            def run(self):
                assert self.a == (1, 2)
                assert self.b == tuple()

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        @luigi_tools.task.copy_params(
            a=luigi_tools.task.ParamRef(TaskListParameter),
            b=luigi_tools.task.ParamRef(TaskListParameter),
        )
        class TaskCopyListParameter(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            def run(self):
                assert self.a == (1, 2)
                assert self.b == tuple()

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with set_luigi_config({"TaskListParameter": {"b": []}, "TaskCopyListParameter": {"b": []}}):
            assert luigi.build([TaskListParameter(), TaskCopyListParameter()], local_scheduler=True)

    @pytest.mark.skipif(
        Version(get_distribution("luigi").version) < Version("3.2"),
        reason="Schemas are only available for luigi>=3.2",
    )
    def test_dict_param_with_schema(self):
        """Test interaction of @copy_params and schemas."""

        class TaskWithListDictParams(luigi.Config):
            """A simple config task."""

            a = luigi.OptionalListParameter(
                description="a in TaskWithListDictParams", schema={"type": "array"}
            )
            b = luigi.OptionalDictParameter(
                description="a in TaskWithListDictParams", schema={"type": "object"}
            )

        @luigi_tools.task.copy_params(
            a=luigi_tools.task.ParamRef(TaskWithListDictParams),
            b=luigi_tools.task.ParamRef(TaskWithListDictParams),
        )
        class TaskCopyListDictParams(luigi_tools.task.GlobalParamMixin, luigi.Task):
            """A simple test task."""

            def run(self):
                assert self.a == (1, 2)
                assert self.b == {"attr1": 1, "attr2": 2}

            def output(self):
                return luigi.LocalTarget("not_existing_file")

        with set_luigi_config(
            {
                "TaskWithListDictParams": {
                    "a": "[1, 2]",
                    "b": json.dumps({"attr1": 1, "attr2": 2}),
                }
            }
        ):
            assert luigi.build([TaskCopyListDictParams()], local_scheduler=True)


class TestForceableTask:
    """Test the rerun feature."""

    def test_no_rerun(self, task_collection):
        """Test that everything is run when all rerun are False and targets are missing."""
        assert all(
            not check_existing_file(i.path) for i in luigi.task.flatten(task_collection.targets)
        )

        assert luigi.build([task_collection.TaskE()], local_scheduler=True)

        assert all(
            check_not_empty_file(i.path) for i in luigi.task.flatten(task_collection.targets)
        )

    def test_no_rerun_with_complete_targets(self, task_collection):
        """Test that nothing is run when all rerun are False and targets are present."""
        for i in luigi.task.flatten(task_collection.targets):
            create_empty_file(i.path)

        assert luigi.build([task_collection.TaskE()], local_scheduler=True)

        assert all(check_empty_file(i.path) for i in luigi.task.flatten(task_collection.targets))

    def test_rerun_with_complete_targets(self, task_collection):
        """Test that everything is run when rerun = True for the root task.

        The targets of all tasks are present and should thus be automatically removed before running
        the workflow.
        """
        for i in luigi.task.flatten(task_collection.targets):
            create_empty_file(i.path)

        with set_luigi_config(
            {
                "TaskA": {"rerun": "true"},
            }
        ):
            assert luigi.build([task_collection.TaskE()], local_scheduler=True)

        assert all(
            check_not_empty_file(i.path) for i in luigi.task.flatten(task_collection.targets)
        )

    def test_rerun_parents_only(self, task_collection):
        """Test that only the parents of the task with rerun = True are run."""
        for i in luigi.task.flatten(task_collection.targets):
            create_empty_file(i.path)

        with set_luigi_config(
            {
                "TaskB": {"rerun": "true"},
            }
        ):
            assert luigi.build([task_collection.TaskE()], local_scheduler=True)

        assert all(
            check_not_empty_file(j.path)
            for task_name, targets in task_collection.targets.items()
            for j in luigi.task.flatten(targets)
            if task_name not in ["TaskA", "TaskC"]
        )
        assert all(
            check_empty_file(j.path)
            for task_name, targets in task_collection.targets.items()
            for j in luigi.task.flatten(targets)
            if task_name in ["TaskA", "TaskC"]
        )

    def test_no_remove_if_recall(self, task_collection, tmpdir):
        """Test that calling a task inside another one does not remove its targets."""

        class TaskF(luigi_tools.task.WorkflowTask):
            """A simple test task."""

            counter = luigi.IntParameter(default=0)
            rerun = luigi.BoolParameter()

            def requires(self):
                return task_collection.TaskE()

            def run(self):
                # Call A inside F but the targets of A should not be removed
                _ = task_collection.TaskA(counter=999)

                for i in luigi.task.flatten(self.output()):
                    create_not_empty_file(i.path)

            def output(self):
                return {
                    "first_target": luigi.LocalTarget(tmpdir / "TaskF.target"),
                    "other_targets": {
                        "second_target": luigi.LocalTarget(tmpdir / "TaskE2.target"),
                        "third_target": luigi.LocalTarget(tmpdir / "TaskE3.target"),
                    },
                }

        for i in luigi.task.flatten(task_collection.targets):
            create_empty_file(i.path)

        with set_luigi_config(
            {
                "TaskB": {"rerun": "true"},
            }
        ):
            assert luigi.build([TaskF()], local_scheduler=True)

        assert all(
            check_not_empty_file(j.path)
            for task_name, targets in task_collection.targets.items()
            for j in luigi.task.flatten(targets)
            if task_name not in ["TaskA", "TaskC"]
        )
        assert all(
            check_empty_file(j.path)
            for task_name, targets in task_collection.targets.items()
            for j in luigi.task.flatten(targets)
            if task_name in ["TaskA", "TaskC"]
        )

    def test_wrapper_task(self, task_collection, tmpdir):
        """Test that it is possible to rerun a wrapper task and all its dependencies."""

        class TaskF(luigi_tools.task.WorkflowWrapperTask):
            """A simple test wrapper task."""

            def requires(self):
                return task_collection.TaskE()

        for i in luigi.task.flatten(task_collection.targets):
            create_empty_file(i.path)

        with set_luigi_config(
            {
                "TaskF": {"rerun": "true"},
            }
        ):
            assert luigi.build([TaskF()], local_scheduler=True)

        assert all(
            check_not_empty_file(j.path)
            for task_name, targets in task_collection.targets.items()
            for j in luigi.task.flatten(targets)
        )


class TestLogTargetMixin:
    """Test the `luigi_tools.task.LogTargetMixin` class."""

    def test_output_logger(self, tmpdir, caplog):
        """Test the logger outputs with one target."""

        class TaskA(luigi_tools.task.LogTargetMixin, luigi.Task):
            """A simple test task."""

            a = luigi.Parameter(default="a")

            def run(self):
                create_empty_file(self.output().path)

            def output(self):
                return luigi.LocalTarget(tmpdir / f"test_{self.a}")

        caplog.clear()
        caplog.set_level(logging.DEBUG)
        assert luigi.build([TaskA()], local_scheduler=True)
        res = [i for i in caplog.record_tuples if i[0] == "luigi_tools.task"]
        assert res == [("luigi_tools.task", 10, f"Output of TaskA task: {tmpdir / 'test_a'}")]

    def test_output_logger_with_dict(self, tmpdir, caplog):
        """Test the logger outputs with a dict target."""

        class AlwaysExistingTarget(luigi.Target):
            """A target that is always considered as existing."""

            def exists(self):
                return True

        class TaskB(luigi_tools.task.LogTargetMixin, luigi.Task):
            """A simple test task."""

            b = luigi.Parameter(default="b")

            def run(self):
                create_empty_file(self.output()["b"].path)

            def output(self):
                return {
                    "b": luigi.LocalTarget(tmpdir / f"test_{self.b}"),
                    "dummy": AlwaysExistingTarget(),
                }

        caplog.clear()
        caplog.set_level(logging.DEBUG)
        assert luigi.build([TaskB()], local_scheduler=True)
        res = [i for i in caplog.record_tuples if i[0] == "luigi_tools.task"]
        assert res == [("luigi_tools.task", 10, f"Output b of TaskB task: {tmpdir / 'test_b'}")]

        class TaskC(luigi.Task):
            """A simple test task."""

            c = luigi.Parameter(default="c")

            def run(self):
                create_empty_file(self.output().path)

            def output(self):
                return luigi.LocalTarget(tmpdir / f"test_{self.c}")

        # Test that other classes do not emit this log entry
        caplog.clear()
        caplog.set_level(logging.DEBUG)
        assert luigi.build([TaskC()], local_scheduler=True)
        res = [i for i in caplog.record_tuples if i[0] == "luigi_tools.task"]
        assert res == []

    def test_output_logger_ignored(self, tmpdir, caplog):
        """Test that there is no logger output when the mixin is located after `luigi.Task`."""

        class TaskD(luigi.Task, luigi_tools.task.LogTargetMixin):
            """A simple test task."""

            d = luigi.Parameter(default="d")

            def run(self):
                create_empty_file(self.output().path)

            def output(self):
                return luigi.LocalTarget(tmpdir / f"test_{self.d}")

        # Test that no log entry is emitted when the mixin is placed after any luigi.Task
        caplog.clear()
        caplog.set_level(logging.DEBUG)
        assert luigi.build([TaskD()], local_scheduler=True)
        res = [i for i in caplog.record_tuples if i[0] == "luigi_tools.task"]
        assert res == []

    def test_output_logger_and_global_param(self, tmpdir, caplog):
        """Test logger outputs when interacting with `luigi_tools.task.GlobalParamMixin`."""

        class TaskE(
            luigi_tools.task.LogTargetMixin,
            luigi_tools.task.GlobalParamMixin,
            luigi.Task,
        ):
            """A simple test task."""

            e = luigi.Parameter(default="e")

            def run(self):
                create_empty_file(self.output().path)

            def output(self):
                return luigi.LocalTarget(tmpdir / f"test_{self.e}")

        # Test that combining several mixins works as intended (i.e. all log entries appear)
        caplog.clear()
        caplog.set_level(logging.DEBUG)
        assert luigi.build([TaskE()], local_scheduler=True)
        res = [i for i in caplog.record_tuples if i[0] == "luigi_tools.task"]
        assert res == [
            (
                "luigi_tools.task",
                10,
                "Attributes of TaskE task after global processing:",
            ),
            ("luigi_tools.task", 10, "Attribute: e == e"),
            ("luigi_tools.task", 10, f"Output of TaskE task: {tmpdir / 'test_e'}"),
        ]


def test_remove_corrupted_output(tmpdir, caplog):
    """Test the `luigi_tools.task.RemoveCorruptedOutputMixin` mixin."""

    class TaskToFail(luigi_tools.task.RemoveCorruptedOutputMixin, luigi.Task):
        """A Task that's expected to fail and create incomplete/wrong output."""

        def output(self):
            return luigi.LocalTarget(tmpdir / "matrix.dat")

        def run(self):
            f_path = self.output().path
            with open(f_path, "w", encoding="utf-8") as f_handle:
                for idx in range(10):
                    f_handle.write(f"{idx}\n")
                    if idx >= 5:
                        raise RuntimeError("something unexpected happened!")

    @TaskToFail.event_handler(luigi.Event.FAILURE)
    def check_exception(task, exception):
        """Gets triggered upon luigi.Event.FAILURE event."""
        assert isinstance(exception, RuntimeError)
        assert exception.args[0] == "something unexpected happened!"

    task_instance = TaskToFail(clean_failed=True)
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    luigi.build([task_instance], local_scheduler=True)

    # there are missing targets
    assert not task_instance.complete()

    res = [i for i in caplog.record_tuples if i[0] == "luigi_tools.util"]
    assert res == [
        ("luigi_tools.util", 10, f"Removing {tmpdir / 'matrix.dat'}"),
    ]

    assert not (tmpdir / "matrix.dat").exists()

    task_instance.clean_failed = False
    luigi.build([task_instance], local_scheduler=True)

    # all targets are produced
    assert task_instance.complete()

    with open(tmpdir / "matrix.dat", "r", encoding="utf-8") as f_handle:
        matrix_values = f_handle.read().splitlines()
        assert len(matrix_values) == 6


class TestCheckUnconsumedParams:
    """Check that warnings are properly emitted when a parameter is not consumed by a task."""

    def test_unconsumed(self, tmpdir, caplog):
        """Simple test with unconsumed parameters."""

        class TaskA(luigi_tools.task.WorkflowTask):
            """A simple test task."""

            a = luigi.Parameter(default="a")

            def run(self):
                create_not_empty_file(self.output().path)

            def output(self):
                return luigi.LocalTarget(tmpdir / f"test_TaskA_{self.a}")

        class TaskB(luigi_tools.task.WorkflowTask):
            """A simple test task."""

            a = luigi.Parameter(default="a")

            def run(self):
                create_not_empty_file(self.output().path)

            def output(self):
                return luigi.LocalTarget(tmpdir / f"test_TaskB_{self.a}")

        with set_luigi_config(
            {
                "TaskA": {
                    "a": "a",
                    "b": "b",
                    "c": "c",
                },
                "TaskB": {
                    "a": "a",
                    "b": "b",
                    "c": "c",
                },
            }
        ):
            with warnings.catch_warnings(record=True) as w:
                warnings.filterwarnings(
                    action="ignore",
                    category=Warning,
                )
                warnings.simplefilter(
                    action="always",
                    category=luigi.parameter.UnconsumedParameterWarning,
                )
                assert luigi.build([TaskA(), TaskB()], local_scheduler=True)

                assert len(w) == 4
                expected = [
                    ("b", "TaskA"),
                    ("c", "TaskA"),
                    ("b", "TaskB"),
                    ("c", "TaskB"),
                ]
                for i, (expected_value, task_name) in zip(w, expected):
                    assert issubclass(i.category, luigi.parameter.UnconsumedParameterWarning)
                    assert str(i.message) == (
                        "The configuration contains the parameter "
                        f"'{expected_value}' with value '{expected_value}' that is not consumed by "
                        f"the task '{task_name}'."
                    )

        assert all(check_not_empty_file(i.path) for i in luigi.task.flatten(TaskA().output()))



# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for luigi-tools targets."""

# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument
import os
import re
import shutil
from pathlib import Path

import luigi
import pytest

import luigi_tools.target
import luigi_tools.task
import luigi_tools.util

from .tools import create_not_empty_file


@pytest.fixture
def reset_prefix():
    """Fixture to reset the prefix of the `luigi_tools.target.OutputLocalTarget` class."""
    assert luigi_tools.target.OutputLocalTarget.get_default_prefix() == Path()
    yield
    luigi_tools.target.OutputLocalTarget.set_default_prefix(None)


class TestOutputTarget:
    """Test the `luigi_tools.target.OutputLocalTarget` class."""

    def test_get_default_prefix(self, tmpdir, reset_prefix):
        """Test the `get_default_prefix` class method."""
        expected = Path()
        assert luigi_tools.target.OutputLocalTarget.get_default_prefix() == expected
        assert (
            luigi_tools.target.OutputLocalTarget("path", create_parent=False).get_default_prefix()
            == expected
        )
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix=tmpdir, create_parent=False
            ).get_default_prefix()
            == expected
        )

        luigi_tools.target.OutputLocalTarget.set_default_prefix(tmpdir / "subdir")

        expected = Path(tmpdir / "subdir")
        assert luigi_tools.target.OutputLocalTarget.get_default_prefix() == expected
        assert (
            luigi_tools.target.OutputLocalTarget("path", create_parent=False).get_default_prefix()
            == expected
        )
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix=tmpdir, create_parent=False
            ).get_default_prefix()
            == expected
        )

    def test_prefix(self, tmpdir, reset_prefix):
        """Test the `get_prefix` method."""
        assert (
            luigi_tools.target.OutputLocalTarget("path", create_parent=False).get_prefix() == Path()
        )
        assert luigi_tools.target.OutputLocalTarget(
            "path", prefix=tmpdir, create_parent=False
        ).get_prefix() == Path(tmpdir)

        luigi_tools.target.OutputLocalTarget.set_default_prefix(tmpdir / "subdir")

        assert luigi_tools.target.OutputLocalTarget(
            "path", create_parent=False
        ).get_prefix() == Path(tmpdir / "subdir")
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix=tmpdir / "test_subdir", create_parent=False
            ).get_prefix()
            == Path(tmpdir) / "test_subdir"
        )
        assert luigi_tools.target.OutputLocalTarget(
            "path", prefix="test_subdir", create_parent=False
        ).get_prefix() == Path("test_subdir")

    def test_pathlib_path(self, tmpdir, reset_prefix):
        """Test the `pathlib_path` property."""
        tmpdir = Path(tmpdir)

        # No default prefix and no prefix
        assert luigi_tools.target.OutputLocalTarget(
            "path", create_parent=False
        ).pathlib_path == Path("path")

        # No default prefix and absolute prefix
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix=tmpdir, create_parent=False
            ).pathlib_path
            == tmpdir / "path"
        )

        # No default prefix and relative prefix
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix="test_subdir", create_parent=False
            ).pathlib_path
            == Path("test_subdir") / "path"
        )

        # Set absolute default prefix
        luigi_tools.target.OutputLocalTarget.set_default_prefix(tmpdir / "absolute_default_subdir")

        # Default prefix and no prefix
        assert (
            luigi_tools.target.OutputLocalTarget("path", create_parent=False).pathlib_path
            == tmpdir / "absolute_default_subdir" / "path"
        )

        # Default prefix and absolute prefix
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix=tmpdir / "other_subdir", create_parent=False
            ).pathlib_path
            == tmpdir / "other_subdir" / "path"
        )

        # Default prefix and relative prefix (replace only the default suffix of the current class)
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix="test_subdir", create_parent=False
            ).pathlib_path
            == Path("test_subdir") / "path"
        )

        # Set relative default prefix
        luigi_tools.target.OutputLocalTarget.set_default_prefix("relative_default_subdir")

        # Default prefix and no prefix
        assert (
            luigi_tools.target.OutputLocalTarget("path", create_parent=False).pathlib_path
            == Path("relative_default_subdir") / "path"
        )

        # Default prefix and absolute prefix
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix=tmpdir / "other_subdir", create_parent=False
            ).pathlib_path
            == tmpdir / "other_subdir" / "path"
        )

        # Default prefix and relative prefix (replace only the default suffix of the current class)
        assert (
            luigi_tools.target.OutputLocalTarget(
                "path", prefix="test_subdir", create_parent=False
            ).pathlib_path
            == Path("test_subdir") / "path"
        )

    def test_in_task(self, tmpdir, reset_prefix):
        """Test the `luigi_tools.target.OutputLocalTarget` class used in an actual `luigi.task`.

        Several tests for the OutputLocalTarget class:

        * using explicit prefix, so the default prefix is ignored
        * using absolute path, so the prefix is ignored
        * using explicit prefix with relative path, so the default prefix is ignored
        * using explicit prefix with absolute path, so the prefix is ignored
        * using default prefix
        """

        class TaskA_OutputLocalTarget(luigi_tools.task.WorkflowTask):
            """A simple test task."""

            def run(self):
                expected = [
                    str(tmpdir / "output_target.test"),
                    str(tmpdir / "absolute_output_target_no_prefix.test"),
                    str(tmpdir / "relative_output_target.test"),
                    str(tmpdir / "subdir" / "output_target_default_prefix.test"),
                    str(tmpdir / "absolute_output_target_prefix.test"),
                ]

                assert [i.path for i in luigi.task.flatten(self.output())] == expected

                for i in luigi.task.flatten(self.output()):
                    os.makedirs(i.pathlib_path.parent, exist_ok=True)
                    create_not_empty_file(i.path)
                    assert i.exists()
                    luigi_tools.util.target_remove(i)
                    assert not i.exists()

            def output(self):
                return [
                    luigi_tools.target.OutputLocalTarget(
                        "output_target.test", prefix=tmpdir, create_parent=False
                    ),
                    luigi_tools.target.OutputLocalTarget(
                        tmpdir / "absolute_output_target_no_prefix.test",
                        create_parent=False,
                    ),
                    luigi_tools.target.OutputLocalTarget(
                        "relative_output_target.test",
                        prefix=tmpdir / "test" / "..",
                        create_parent=False,
                    ),
                    luigi_tools.target.OutputLocalTarget(
                        "output_target_default_prefix.test", create_parent=False
                    ),
                    luigi_tools.target.OutputLocalTarget(
                        tmpdir / "absolute_output_target_prefix.test",
                        prefix=tmpdir / "test",
                        create_parent=False,
                    ),
                ]

        luigi_tools.target.OutputLocalTarget.set_default_prefix(tmpdir / "subdir")

        assert luigi.build([TaskA_OutputLocalTarget()], local_scheduler=True)

        luigi_tools.target.OutputLocalTarget.set_default_prefix(None)
        target = luigi_tools.target.OutputLocalTarget("test", prefix=None, create_parent=False)
        assert target.path == "test"
        luigi_tools.target.OutputLocalTarget.set_default_prefix(None)
        assert luigi_tools.target.OutputLocalTarget.get_default_prefix().absolute() == Path(
            os.getcwd()
        )

        other_subdir = tmpdir / "other_subdir"
        luigi_tools.target.OutputLocalTarget.set_default_prefix(other_subdir / "test" / "create")
        target = luigi_tools.target.OutputLocalTarget("test_file", prefix=None, create_parent=False)

        target.mkdir()
        assert target.pathlib_path.parent.is_dir()
        assert not target.pathlib_path.exists()
        shutil.rmtree(other_subdir)

        target.mkdir(is_dir=True)
        assert target.pathlib_path.is_dir()

        with pytest.raises(OSError):
            target.mkdir(exist_ok=False)

        shutil.rmtree(other_subdir)
        with pytest.raises(FileNotFoundError):
            target.mkdir(parents=False)

        auto_target = luigi_tools.target.OutputLocalTarget("test_file", create_parent=True)
        assert auto_target.pathlib_path.parent.is_dir()
        assert not auto_target.pathlib_path.exists()

        not_auto_target = luigi_tools.target.OutputLocalTarget(
            "test_dir/test_file", create_parent=False
        )
        assert not not_auto_target.pathlib_path.parent.exists()
        assert not not_auto_target.pathlib_path.exists()

        class NoPrefix(luigi_tools.target.OutputLocalTarget):
            """A target with no prefix."""

        assert NoPrefix("path").path == other_subdir / "test" / "create" / "path"

    def test_super_prefix(self, tmpdir, reset_prefix):
        """Test the super prefix to build a target hierarchy."""

        # pylint: disable=useless-super-delegation
        class SubOutputLocalTarget(luigi_tools.target.OutputLocalTarget):
            """A simple target with a `str` prefix."""

            __prefix = "sub_prefix"  # Use a string

            def __init__(self, *args, prefix=None, create_parent=False, **kwargs):
                super().__init__(*args, prefix=prefix, create_parent=create_parent, **kwargs)

        class SubSubOutputLocalTarget(SubOutputLocalTarget):
            """A simple target with a `pathlib.Path` prefix."""

            __prefix = Path("sub_sub_prefix")  # Use a pathlib.Path

            def __init__(self, *args, prefix=None, create_parent=False, **kwargs):
                super().__init__(*args, prefix=prefix, create_parent=create_parent, **kwargs)

        class NoPrefix(SubSubOutputLocalTarget):
            """Test that a target without prefix in the MRO does not break the feature."""

            def __init__(self, *args, prefix=None, create_parent=False, **kwargs):
                super().__init__(*args, prefix=prefix, create_parent=create_parent, **kwargs)

        class SubSubSubOutputLocalTarget(NoPrefix):
            """A simple target with another sub-prefix."""

            __prefix = Path("sub_sub_sub_prefix")

            def __init__(self, *args, prefix=None, create_parent=False, **kwargs):
                super().__init__(*args, prefix=prefix, create_parent=create_parent, **kwargs)

        assert luigi_tools.target.OutputLocalTarget("path", create_parent=False).path == "path"

        # Test with default prefixes
        assert SubOutputLocalTarget("path").path == "sub_prefix/path"
        assert SubSubOutputLocalTarget("path").path == "sub_prefix/sub_sub_prefix/path"
        assert (
            SubSubSubOutputLocalTarget("path").path
            == "sub_prefix/sub_sub_prefix/sub_sub_sub_prefix/path"
        )

        # Test with new relative path for parent
        SubOutputLocalTarget.set_default_prefix("new_sub_prefix")
        assert SubOutputLocalTarget("path").path == "new_sub_prefix/path"
        assert SubSubOutputLocalTarget("path").path == "new_sub_prefix/sub_sub_prefix/path"
        assert (
            SubSubSubOutputLocalTarget("path").path
            == "new_sub_prefix/sub_sub_prefix/sub_sub_sub_prefix/path"
        )

        # Test with new absolute path for parent
        SubOutputLocalTarget.set_default_prefix("/tmp/sub_prefix")
        assert SubOutputLocalTarget("path").path == "/tmp/sub_prefix/path"
        assert SubSubOutputLocalTarget("path").path == "/tmp/sub_prefix/sub_sub_prefix/path"
        assert (
            SubSubSubOutputLocalTarget("path").path
            == "/tmp/sub_prefix/sub_sub_prefix/sub_sub_sub_prefix/path"
        )

        # Test with relative paths for both classes
        SubOutputLocalTarget.set_default_prefix("sub_prefix_2")
        SubSubOutputLocalTarget.set_default_prefix("sub_sub_prefix_2")
        assert SubOutputLocalTarget("path").path == "sub_prefix_2/path"
        assert SubSubOutputLocalTarget("path").path == "sub_prefix_2/sub_sub_prefix_2/path"
        assert (
            SubSubSubOutputLocalTarget("path").path
            == "sub_prefix_2/sub_sub_prefix_2/sub_sub_sub_prefix/path"
        )

        # Test with absolute paths for parent class and relative path for child
        SubOutputLocalTarget.set_default_prefix("/tmp/sub_prefix_2")
        SubSubOutputLocalTarget.set_default_prefix("sub_sub_prefix_2")
        assert SubOutputLocalTarget("path").path == "/tmp/sub_prefix_2/path"
        assert SubSubOutputLocalTarget("path").path == "/tmp/sub_prefix_2/sub_sub_prefix_2/path"
        assert (
            SubSubSubOutputLocalTarget("path").path
            == "/tmp/sub_prefix_2/sub_sub_prefix_2/sub_sub_sub_prefix/path"
        )

        # Test with absolute paths for both parent and child classes
        SubOutputLocalTarget.set_default_prefix("/tmp/sub_prefix_2")
        SubSubOutputLocalTarget.set_default_prefix("/tmp/sub_sub_prefix_2")
        assert SubOutputLocalTarget("path").path == "/tmp/sub_prefix_2/path"
        assert SubSubOutputLocalTarget("path").path == "/tmp/sub_sub_prefix_2/path"
        assert (
            SubSubSubOutputLocalTarget("path").path
            == "/tmp/sub_sub_prefix_2/sub_sub_sub_prefix/path"
        )

        # Reset prefix
        SubSubOutputLocalTarget.set_default_prefix(None)
        assert SubOutputLocalTarget("path").path == "/tmp/sub_prefix_2/path"
        assert SubSubOutputLocalTarget("path").path == "/tmp/sub_prefix_2/path"
        assert (
            SubSubSubOutputLocalTarget("path").path == "/tmp/sub_prefix_2/sub_sub_sub_prefix/path"
        )

        # Reset parent prefix
        SubOutputLocalTarget.set_default_prefix(None)
        assert SubOutputLocalTarget("path").path == "path"
        assert SubSubOutputLocalTarget("path").path == "path"
        assert SubSubSubOutputLocalTarget("path").path == "sub_sub_sub_prefix/path"

    @pytest.fixture
    def reset_target_no_prefix(self):
        """Fixture to reset the prefix of the `luigi_tools.target.OutputLocalTarget` class."""
        luigi_tools.target.OutputLocalTarget.set_default_prefix(None)
        yield
        luigi_tools.target.OutputLocalTarget.set_default_prefix(None)

    def test_child_classes(self, reset_target_no_prefix):
        """Test a child class without any default prefix."""

        class TestTarget(luigi_tools.target.OutputLocalTarget):
            """A simple test task."""

            def __init__(self, *args, prefix=None, create_parent=False, **kwargs):
                super().__init__(*args, prefix=prefix, create_parent=create_parent, **kwargs)

        test_target = TestTarget("a_path")
        assert test_target.path == "a_path"

        test_target = TestTarget("a_path", prefix="instance_prefix")
        assert test_target.path == "instance_prefix/a_path"

    @pytest.fixture
    def reset_target_relative_prefix(self):
        """Fixture to reset the prefix of the `luigi_tools.target.OutputLocalTarget` class."""
        luigi_tools.target.OutputLocalTarget.set_default_prefix("parent_default_prefix")
        yield
        luigi_tools.target.OutputLocalTarget.set_default_prefix(None)

    def test_child_classes_with_parent_prefix(self, reset_target_relative_prefix):
        """Test a child class when a default prefix is given to OutputLocalTarget."""
        # pylint: disable=useless-super-delegation

        class TestTarget(luigi_tools.target.OutputLocalTarget):
            """A simple test task."""

            def __init__(self, *args, prefix=None, create_parent=False, **kwargs):
                super().__init__(*args, prefix=prefix, create_parent=create_parent, **kwargs)

        class TestTargetChild(TestTarget):
            """A simple test task."""

            def __init__(self, *args, prefix=None, create_parent=False, **kwargs):
                super().__init__(*args, prefix=prefix, create_parent=create_parent, **kwargs)

        test_target = TestTarget("a_path")
        assert test_target.path == "parent_default_prefix/a_path"

        test_target = TestTarget("a_path", prefix="instance_prefix")
        assert test_target.path == "parent_default_prefix/instance_prefix/a_path"

        # Add a default prefix to TestTarget
        TestTarget.set_default_prefix("default_prefix")

        test_target = TestTarget("a_path")
        assert test_target.path == "parent_default_prefix/default_prefix/a_path"

        test_target = TestTarget("a_path", prefix="instance_prefix")
        assert test_target.path == "parent_default_prefix/instance_prefix/a_path"

        # Add a default prefix to TestTargetChild
        test_target = TestTargetChild("a_path")
        assert test_target.path == "parent_default_prefix/default_prefix/a_path"

        test_target = TestTargetChild("a_path", prefix="child_instance_prefix")
        assert (
            test_target.path == "parent_default_prefix/default_prefix/child_instance_prefix/a_path"
        )

        TestTargetChild.set_default_prefix("child_default_prefix")

        test_target = TestTargetChild("a_path")
        assert (
            test_target.path == "parent_default_prefix/default_prefix/child_default_prefix/a_path"
        )

        test_target = TestTargetChild("a_path", prefix="child_instance_prefix")
        assert (
            test_target.path == "parent_default_prefix/default_prefix/child_instance_prefix/a_path"
        )

        # Remove the default prefix of TestTarget
        TestTarget.set_default_prefix(None)

        test_target = TestTargetChild("a_path")
        assert test_target.path == "parent_default_prefix/child_default_prefix/a_path"

        test_target = TestTargetChild("a_path", prefix="child_instance_prefix")
        assert test_target.path == "parent_default_prefix/child_instance_prefix/a_path"

        # Remove the default prefix of TestTargetChild
        TestTargetChild.set_default_prefix(None)

        test_target = TestTargetChild("a_path")
        assert test_target.path == "parent_default_prefix/a_path"

        test_target = TestTargetChild("a_path", prefix="child_instance_prefix")
        assert test_target.path == "parent_default_prefix/child_instance_prefix/a_path"

    def test_repr_and_str(self):
        """Test the __repr__ and __str__ methods."""
        assert str(luigi_tools.target.OutputLocalTarget("path", create_parent=False)) == "path"
        assert (
            str(
                luigi_tools.target.OutputLocalTarget(
                    "path", prefix="test_prefix", create_parent=False
                )
            )
            == "test_prefix/path"
        )

        assert (
            re.match(
                r"<OutputLocalTarget at 0x\S+; path>",
                repr(luigi_tools.target.OutputLocalTarget("path", create_parent=False)),
            )
            is not None
        )
        assert (
            re.match(
                r"<OutputLocalTarget at 0x\S+; test_prefix/path>",
                repr(
                    luigi_tools.target.OutputLocalTarget(
                        "path", prefix="test_prefix", create_parent=False
                    )
                ),
            )
            is not None
        )

        try:
            luigi_tools.target.OutputLocalTarget.set_default_prefix("default_prefix")

            assert (
                re.match(
                    r"<OutputLocalTarget at 0x\S+; default_prefix/path>",
                    repr(luigi_tools.target.OutputLocalTarget("path", create_parent=False)),
                )
                is not None
            )
            assert (
                re.match(
                    r"<OutputLocalTarget at 0x\S+; test_prefix/path>",
                    repr(
                        luigi_tools.target.OutputLocalTarget(
                            "path", prefix="test_prefix", create_parent=False
                        )
                    ),
                )
                is not None
            )
        finally:
            luigi_tools.target.OutputLocalTarget.set_default_prefix(None)



# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides some specific luigi tasks and associated tools."""
import logging
import types
from copy import deepcopy

import luigi

from luigi_tools.util import apply_over_outputs
from luigi_tools.util import apply_over_required
from luigi_tools.util import recursive_check
from luigi_tools.util import target_remove

L = logging.getLogger(__name__)


_no_default_value = "__no_default_value__"


class DuplicatedParameterError(Exception):
    """Exception raised when a parameter is duplicated in a task."""


class GlobalParameterNoValueError(Exception):
    """Exception raised when the value of a global parameter can not be found."""


class RerunMixin:
    """Mixin used to force a task to run again by setting the 'rerun' parameter to True."""

    rerun = luigi.BoolParameter(
        significant=False,
        default=False,
        description="Trigger to force the task to rerun.",
    )  #:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if recursive_check(self):
            apply_over_outputs(self, target_remove)

        if isinstance(self, luigi.WrapperTask) and self.rerun:

            def recursive_target_remove(task, *args, **kwargs):
                apply_over_outputs(task, target_remove)
                apply_over_required(task, recursive_target_remove)

            recursive_target_remove(self)


class LogTargetMixin:
    """Mixin used to log output paths of this task.

    .. note::
        To have an effect, this mixin must be placed to the left of the first class
        inheriting from :class:`luigi.Task` in the base class list.
    """

    def __init__(self, *args, **kwargs):
        event_handler = super().event_handler

        # pylint: disable=unused-variable
        @event_handler(luigi.Event.SUCCESS)
        def log_targets(task):
            """Hook to log output targets of a task."""

            def log_func(target, key=None):
                if not hasattr(target, "path"):
                    return
                class_name = task.__class__.__name__
                if key is None:
                    L.debug("Output of %s task: %s", class_name, target.path)
                else:
                    L.debug("Output %s of %s task: %s", key, class_name, target.path)

            apply_over_outputs(task, log_func)

        super().__init__(*args, **kwargs)


class GlobalParamMixin:
    """Mixin used to add customisable global parameters.

    For the tasks that inherit from this GlobalParamMixin, when a parameter is linked to
    another one and it value is None, its value is automatically replaced by the value
    of the linked parameter.
    See :class:`copy_params` for details about parameter linking.
    """

    def __init__(self, *args, **kwargs):
        event_handler = super().event_handler

        # pylint: disable=unused-variable
        @event_handler(luigi.Event.START)
        def log_parameters(self):
            """Hook to log actual parameter values considering their global processing."""
            class_name = self.__class__.__name__
            L.debug("Attributes of %s task after global processing:", class_name)
            for name in self.get_param_names():
                try:
                    L.debug("Attribute: %s == %s", name, getattr(self, name))
                except Exception:  # pylint: disable=broad-except; # pragma: no cover
                    L.debug("Can't print '%s' attribute for unknown reason", name)

        super().__init__(*args, **kwargs)

    def __getattribute__(self, name):
        """Get the attribute value and get it from the linked parameter is the value is None."""
        tmp = super().__getattribute__(name)
        if not isinstance(tmp, str) or tmp != _no_default_value:
            return tmp
        if hasattr(self, "_global_params"):
            global_param = self._global_params.get(name)
            if global_param is not None:
                return getattr(global_param.cls(), global_param.name)
        raise GlobalParameterNoValueError(
            f"The value of the '{name}' parameter can not be retrieved from the linked parameter."
        )

    def __setattr__(self, name, value):
        """Send a warning logger entry when a global parameter is set to None."""
        try:
            global_params = self._global_params
        except AttributeError:
            global_params = {}
        if name in global_params and isinstance(value, str) and value == _no_default_value:
            L.warning(
                "The Parameter '%s' of the task '%s' is set to None, thus the global "
                "value will be taken frow now on",
                name,
                self.__class__.__name__,
            )
        return super().__setattr__(name, value)


class RemoveCorruptedOutputMixin:
    """Mixin used to remove incomplete outputs of a failed Task.

    The clean_failed parameter must be set True to enable this feature.
    """

    clean_failed = luigi.BoolParameter(
        significant=False,
        default=False,
        description="Trigger to remove the outputs of the failed tasks.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        event_handler = super().event_handler

        # pylint: disable=unused-variable, unused-argument
        @event_handler(luigi.Event.FAILURE)
        def remove_all_output(self, exception):
            class_name = self.__class__.__name__
            if self.clean_failed:
                L.debug("%s failed! Cleaning the targets...", class_name)
                apply_over_outputs(self, target_remove)


class ParamRef:
    """Class to store parameter reference information."""

    def __init__(self, cls, name=None, default=_no_default_value):
        self.cls = cls
        self.name = name
        self.default = default


class copy_params:
    """Copy a parameter from another Task.

    This decorator takes kwargs where keys are the parameter names and the values are
    :class:`ParamRef` instances.

    If no default value is given to the :class:`ParamRef`, two behaviours are possible:

        * If the task inherits from the :class:`GlobalParamMixin`, the parameter is linked to the
          one of the base class. The default parameter value is then the actual value of the
          linked parameter.
        * If the task does not inherit from the :class:`GlobalParamMixin`, the default value is
          copied from the linked parameter.

    Differences with :class:`luigi.util.inherits`:

        * :class:`luigi.util.inherits` set all other class' parameters that the current one is
          missing. It is not possible to select a subset of parameters.
        * :class:`luigi.util.inherits` set the same value to the other class' parameter than the
          current one's parameter. This class does not set the same value, except for global
          parameters with no given value.

    **Usage**:

    .. code-block:: python

        class AnotherTask(luigi.Task):
            m = luigi.IntParameter(default=1)

        @copy_params(m=ParamRef(AnotherTask))
        class MyFirstTask(luigi.Task):
            def run(self):
               print(self.m) # this will be defined and print 1
               # ...

        @copy_params(another_m=ParamRef(AnotherTask, "m"))
        class MySecondTask(luigi.Task):
            def run(self):
               print(self.another_m) # this will be defined and print 1
               # ...

        @copy_params(another_m=ParamRef(AnotherTask, "m", 5))
        class MyFirstTask(luigi.Task):
            def run(self):
               print(self.another_m) # this will be defined and print 5
               # ...

        @copy_params(another_m=ParamRef(AnotherTask, "m"))
        class MyFirstTask(GlobalParamMixin, luigi.Task):
            def run(self):
               # this will be defined and print 1 if self.another_m is not explicitly set
               # (this means that self.another_m == luigi_tools.tasks._no_default_value)
               print(self.another_m)
               # ...

    .. warning::
        There is a limitation of using :class:`copy_params` on a :class:`GlobalParamMixin`.
        In this particular case, the task from which a parameter is copied must be called
        before the inheriting task (which does not mean it must be executed before, just
        ``task()`` is enough). This is due to the metaclass magic of luigi.
    """

    def __init__(self, **params_to_copy):
        """Initialize the decorator."""
        super().__init__()
        if not params_to_copy:
            raise ValueError("params_to_copy cannot be empty")

        self.params_to_copy = params_to_copy

    def __call__(self, task_that_inherits):
        """Set and configure the attributes."""
        # Get all parameters
        for param_name, attr in self.params_to_copy.items():
            # Check if the parameter exists in the inheriting task
            if not hasattr(task_that_inherits, param_name):
                if attr.name is None:
                    attr.name = param_name

                # Copy param
                param = getattr(attr.cls, attr.name)
                new_param = deepcopy(param)

                # Set default value if required
                if attr.default != _no_default_value:
                    new_param._default = attr.default
                elif issubclass(task_that_inherits, GlobalParamMixin):
                    new_param._default = _no_default_value
                elif param._default == luigi.parameter._no_value:
                    # The deepcopy make the new_param._default != luigi.parameter._no_value
                    # so we must reset it in this case
                    new_param._default = luigi.parameter._no_value

                # Add it to the inheriting task with new default values
                setattr(task_that_inherits, param_name, new_param)

                # Do not emit warning because of wrong type
                def _warn_on_wrong_global_param_type(_self, global_param_name, global_param_value):
                    if global_param_value != _no_default_value:
                        # pylint: disable=protected-access
                        _self._actual_warn_on_wrong_param_type(
                            global_param_name, global_param_value
                        )

                new_param._actual_warn_on_wrong_param_type = new_param._warn_on_wrong_param_type
                new_param._warn_on_wrong_param_type = types.MethodType(
                    _warn_on_wrong_global_param_type, new_param
                )

                # Fix normalize() to not normalize when the parameter does not have default value
                def normalize(_self, value):
                    # pylint: disable=protected-access
                    if value == _no_default_value:
                        return value
                    else:
                        return _self._actual_normalize(value)

                new_param._actual_normalize = new_param.normalize
                new_param.normalize = types.MethodType(normalize, new_param)

                # Add link to global parameter
                if issubclass(task_that_inherits, GlobalParamMixin):
                    if not hasattr(task_that_inherits, "_global_params"):
                        task_that_inherits._global_params = {}
                    task_that_inherits._global_params[param_name] = attr
            else:
                raise DuplicatedParameterError(
                    f"The parameter '{param_name}' already exists in the task "
                    f"'{task_that_inherits.__class__.__name__}'."
                )

        return task_that_inherits


class WorkflowTask(GlobalParamMixin, RerunMixin, luigi.Task):
    """Default task used in workflows.

    This task can be forced running again by setting the 'rerun' parameter to True.
    It can also use copy and link parameters from other tasks.
    """


class WorkflowWrapperTask(WorkflowTask, luigi.WrapperTask):
    """Base wrapper class with global parameters."""



"""Configuration file for the Sphinx documentation builder."""

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from luigi import Parameter
from pkg_resources import get_distribution

from luigi_tools.util import _param_repr

sys.path.insert(0, os.path.abspath("_ext"))

# -- Project information -----------------------------------------------------

project = "luigi-tools"

# The short X.Y version
version = get_distribution("luigi-tools").version

# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "m2r2",
    "luigi_move",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx-bluebrain-theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

html_theme_options = {
    "metadata_distribution": "luigi-tools",
}

html_title = "Luigi tools"

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# autosummary settings
autosummary_generate = True

# autodoc settings
autodoc_typehints = "signature"
autodoc_default_options = {"members": True, "show-inheritance": True}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "graphviz": ("https://graphviz.readthedocs.io/en/stable", None),
    "luigi": ("https://luigi.readthedocs.io/en/stable", None),
}

# pylint: disable=protected-access
Parameter.__repr__ = lambda self: _param_repr(self.description, self._default)



# Copyright 2021 Blue Brain Project / EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides some specific luigi parameters."""
import collections.abc
import dataclasses
import typing

import luigi
import typing_extensions
from luigi.freezing import FrozenOrderedDict
from luigi.parameter import OptionalParameterMixin


class ExtParameter(luigi.Parameter):
    """Class to parse file extension parameters."""

    def normalize(self, x):
        """Normalize the given value to an extension suffix without dot."""
        if x.startswith("."):
            x = x[1:]
        return x


class RatioParameter(luigi.NumericalParameter):
    """Class to parse ratio parameters.

    The argument must be a float between 0 and 1.
    The operators to include or exclude the boundaries can be set with 'left_op' and
    'right_op' parameters. Defaults operators include the boundaries.
    """

    def __init__(
        self,
        *args,
        left_op=luigi.parameter.operator.le,
        right_op=luigi.parameter.operator.le,
        **kwargs,
    ):
        super().__init__(
            *args,
            min_value=0,
            max_value=1,
            var_type=float,
            left_op=left_op,
            right_op=right_op,
            **kwargs,
        )

    def normalize(self, x):
        """Ensure that the value is contained in the given interval.

        The ``parse()`` method is not called when the parameters are given to the constructor,
        but the value check is only done there. This ``normalize()`` method ensures that the
        interval check is made even when for arguments given to the constructor.
        """
        if x is not None:
            return super().parse(x)
        else:
            return x


class BoolParameter(luigi.BoolParameter):
    """Class to parse boolean parameters and set explicit parsing when default is True."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._default is True:
            if kwargs.get("parsing", "explicit") == self.__class__.IMPLICIT_PARSING:
                raise ValueError(
                    "A BoolParameter with 'default = True' can not use implicit parsing."
                )
            self.parsing = self.__class__.EXPLICIT_PARSING


class OptionalBoolParameter(OptionalParameterMixin, BoolParameter):
    """Class to parse optional bool parameters."""

    expected_type = bool


class OptionalRatioParameter(OptionalParameterMixin, RatioParameter):
    """Class to parse optional ratio parameters."""

    expected_type = float


class DataclassParameter(luigi.DictParameter):
    """Class to parse, serialize, and normalize nested dataclasses.

    Note: Similar to DictParameter, json serialization transforms int keys of dicts into str. The
    dataclass however transforms the keys into their correct type if the respective annotation is
    available.
    """

    def __init__(self, cls_type, *args, **kwargs):
        if not _is_dataclass(cls_type):
            raise TypeError(f"Class type {cls_type.__name__!r} is not a dataclass.")

        self._cls_type = cls_type
        super().__init__(*args, **kwargs)

    def normalize(self, value):
        """Normalize the given value to a dataclass initialized object."""
        if isinstance(value, self._cls_type):
            value = dataclasses.asdict(value)
        return _instantiate(self._cls_type, value)

    def serialize(self, x):
        """Serialize a dataclass object."""
        data = dataclasses.asdict(x)
        return super().serialize(data)


def _instantiate(cls, data):
    # pylint: disable=too-many-return-statements

    if data is None:
        return None

    if cls in {str, int, float, bool}:
        return cls(data)

    if _is_optional(cls):
        return _instantiate(typing_extensions.get_args(cls)[0], data)

    if _is_dataclass(cls):
        return _instantiate_dataclass(cls, data, _instantiate)

    if _is_sequence(data):
        return _instantiate_sequence(cls, data, _instantiate)

    if _is_mapping(data):
        return _instantiate_mapping(cls, data, _instantiate)

    return data


def _is_dataclass(cls):
    return isinstance(cls, type) and dataclasses.is_dataclass(cls)


def _instantiate_dataclass(cls, data, func):
    """Instantiate a dataclass from the given data using the instantiation func for recursion."""
    args = {f.name: func(f.type, data[f.name]) for f in dataclasses.fields(cls)}
    return cls(**args)


def _is_sequence(data):
    return isinstance(data, collections.abc.Sequence) and not isinstance(data, str)


def _is_optional(cls):
    """Optional types have Union as origin and args of the form [Type, NoneType].

    In that case the arguments of the Type are returned, if any.
    """
    return typing_extensions.get_origin(cls) is typing.Union and type(
        None
    ) in typing_extensions.get_args(cls)


def _get_type_args(cls):
    return typing_extensions.get_args(cls) or []


def _instantiate_sequence(cls, data, func):
    args = _get_type_args(cls)

    # Use type annotation to cast the sequence values
    if args:
        return tuple(func(args[0], v) for v in data)

    # Otherwise the types from the data
    return tuple(func(type(v), v) for v in data)


def _is_mapping(data):
    return isinstance(data, collections.abc.Mapping)


def _instantiate_mapping(cls, data, func):
    args = _get_type_args(cls)

    # Use key, value type annotations for casting
    if args:
        assert len(args) == 2, args
        arguments_generator = (
            (
                func(args[0], k),
                func(args[1], v),
            )
            for k, v in data.items()
        )
    # Otherwise use the types from the data
    else:
        arguments_generator = (
            (
                func(type(k), k),
                func(type(v), v),
            )
            for k, v in data.items()
        )
    return FrozenOrderedDict(arguments_generator)
